@@ -4,6 +4,7 @@
  *
  * Copyright © 1994,1995 Ian Jackson <ian@chiark.greenend.org.uk>
  * Copyright © 2000,2001 Wichert Akkerman
+ * Copyright © 2006-2012 Guillem Jover <guillem@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -30,6 +31,7 @@
 #include <dpkg/macros.h>
 #include <dpkg/varbuf.h>
 #include <dpkg/version.h>
+#include <dpkg/arch.h>
 
 DPKG_BEGIN_DECLS
 
@@ -68,10 +70,12 @@ struct dependency {
 
 struct deppossi {
   struct dependency *up;
-  struct pkginfo *ed;
+  struct pkgset *ed;
   struct deppossi *next, *rev_next, *rev_prev;
+  const struct dpkg_arch *arch;
   struct versionrevision version;
   enum depverrel verrel;
+  bool arch_is_implicit;
   bool cyclebreak;
 };
 
@@ -105,13 +109,21 @@ struct filedetails {
  */
 struct pkgbin {
   struct dependency *depends;
-  struct deppossi *depended;
   /* The ‘essential’ flag, true = yes, false = no (absent). */
   bool essential;
+  enum pkgmultiarch {
+    multiarch_no,
+    multiarch_same,
+    multiarch_allowed,
+    multiarch_foreign,
+  } multiarch;
+  const struct dpkg_arch *arch;
+  /* The following is the "pkgname:archqual" cached string, if this was a
+   * C++ class this member would be mutable. */
+  const char *pkgname_archqual;
   const char *description;
   const char *maintainer;
   const char *source;
-  const char *arch;
   const char *installedsize;
   const char *origin;
   const char *bugs;
@@ -154,8 +166,9 @@ struct perpackagestate;
  * Note: Usually referred in the code as pig.
  */
 struct pkginfo {
-  struct pkginfo *next;
-  const char *name;
+  struct pkgset *set;
+  struct pkginfo *arch_next;
+
   enum pkgwant {
     want_unknown, want_install, want_hold, want_deinstall, want_purge,
     /* Not allowed except as special sentinel value in some places. */
@@ -202,12 +215,28 @@ struct pkginfo {
   struct trigpend *trigpend_head;
 };
 
+/**
+ * Node describing a package set sharing the same package name.
+ */
+struct pkgset {
+  struct pkgset *next;
+  const char *name;
+  struct pkginfo pkg;
+  struct {
+    struct deppossi *available;
+    struct deppossi *installed;
+  } depended;
+  int installed_instances;
+};
+
 /*** from dbdir.c ***/
 
 const char *dpkg_db_set_dir(const char *dir);
 const char *dpkg_db_get_dir(void);
 char *dpkg_db_get_path(const char *pathpart);
 
+#include <dpkg/atomic-file.h>
+
 /*** from dbmodify.c ***/
 
 enum modstatdb_rw {
@@ -230,6 +259,7 @@ bool modstatdb_can_lock(void);
 void modstatdb_lock(void);
 void modstatdb_unlock(void);
 enum modstatdb_rw modstatdb_open(enum modstatdb_rw reqrwflags);
+enum modstatdb_rw modstatdb_get_status(void);
 void modstatdb_note(struct pkginfo *pkg);
 void modstatdb_note_ifwrite(struct pkginfo *pkg);
 void modstatdb_checkpoint(void);
@@ -237,16 +267,25 @@ void modstatdb_shutdown(void);
 
 /*** from database.c ***/
 
+void pkgset_blank(struct pkgset *set);
+int pkgset_installed_instances(struct pkgset *set);
+
 void pkg_blank(struct pkginfo *pp);
 void pkgbin_blank(struct pkgbin *pifp);
 bool pkg_is_informative(struct pkginfo *pkg, struct pkgbin *info);
 
-struct pkginfo *pkg_db_find(const char *name);
-int pkg_db_count(void);
+struct pkgset *pkg_db_find_set(const char *name);
+struct pkginfo *pkg_db_get_singleton(struct pkgset *set);
+struct pkginfo *pkg_db_find_singleton(const char *name);
+struct pkginfo *pkg_db_get_pkg(struct pkgset *set, const struct dpkg_arch *arch);
+struct pkginfo *pkg_db_find_pkg(const char *name, const struct dpkg_arch *arch);
+int pkg_db_count_set(void);
+int pkg_db_count_pkg(void);
 void pkg_db_reset(void);
 
 struct pkgiterator *pkg_db_iter_new(void);
-struct pkginfo *pkg_db_iter_next(struct pkgiterator *iter);
+struct pkgset *pkg_db_iter_next_set(struct pkgiterator *iter);
+struct pkginfo *pkg_db_iter_next_pkg(struct pkgiterator *iter);
 void pkg_db_iter_free(struct pkgiterator *iter);
 
 void pkg_db_report(FILE *);
@@ -268,9 +307,17 @@ enum parsedbflags {
   pdb_lax_version_parser = 040,
   /* Perform laxer parsing, used to transition to stricter parsing. */
   pdb_lax_parser = pdb_lax_version_parser,
+  /* Parse the control file from a binary .deb package. */
+  pdb_deb_control = 0100,
+
+  /* Standard operations. */
+  pdb_parse_status = pdb_lax_parser | pdb_weakclassification,
+  pdb_parse_update = pdb_parse_status | pdb_deb_control,
+  pdb_parse_available = pdb_recordavailable | pdb_rejectstatus | pdb_lax_parser,
+  pdb_parse_binary = pdb_recordavailable | pdb_rejectstatus | pdb_deb_control,
 };
 
-const char *pkg_name_is_illegal(const char *p, const char **ep);
+const char *pkg_name_is_illegal(const char *p);
 int parsedb(const char *filename, enum parsedbflags, struct pkginfo **donep);
 void copy_dependency_links(struct pkginfo *pkg,
                            struct dependency **updateme,
@@ -282,6 +329,7 @@ void copy_dependency_links(struct pkginfo *pkg,
 #include <dpkg/namevalue.h>
 
 extern const struct namevalue booleaninfos[];
+extern const struct namevalue multiarchinfos[];
 extern const struct namevalue priorityinfos[];
 extern const struct namevalue statusinfos[];
 extern const struct namevalue eflaginfos[];
@@ -297,6 +345,24 @@ int parseversion(struct versionrevision *rversion, const char *,
 const char *versiondescribe(const struct versionrevision*,
                             enum versiondisplayepochwhen);
 
+enum pkg_name_arch_when {
+  /* Never display arch. */
+  pnaw_never,
+  /* Display arch only when it's non-ambiguous. */
+  pnaw_nonambig,
+  /* Display arch only when it's a foreign one. */
+  pnaw_foreign,
+  /* Always display arch. */
+  pnaw_always,
+};
+
+void varbuf_add_pkgbin_name(struct varbuf *vb, const struct pkginfo *pkg,
+                            const struct pkgbin *pkgbin,
+                            enum pkg_name_arch_when pnaw);
+const char *pkgbin_name(struct pkginfo *pkg, struct pkgbin *pkgbin,
+                        enum pkg_name_arch_when pnaw);
+const char *pkg_name(struct pkginfo *pkg, enum pkg_name_arch_when pnaw);
+
 /*** from dump.c ***/
 
 void writerecord(FILE*, const char*,
@@ -319,7 +385,6 @@ void varbufdependency(struct varbuf *vb, struct dependency *dep);
 
 /*** from vercmp.c ***/
 
-bool versionsatisfied(struct pkgbin *it, struct deppossi *against);
 bool versionsatisfied3(const struct versionrevision *it,
                        const struct versionrevision *ref,
                        enum depverrel verrel);
@@ -328,6 +393,11 @@ int versioncompare(const struct versionrevision *version,
 bool epochsdiffer(const struct versionrevision *a,
                   const struct versionrevision *b);
 
+/*** from depcon.c ***/
+
+bool versionsatisfied(struct pkgbin *it, struct deppossi *against);
+bool archsatisfied(struct pkgbin *it, struct deppossi *against);
+
 /*** from nfmalloc.c ***/
 void *nfmalloc(size_t);
 char *nfstrsave(const char*);
