@@ -4,6 +4,10 @@
  *
  * Copyright © 1995 Ian Jackson <ian@chiark.greenend.org.uk>
  * Copyright © 2001 Wichert Akkerman
+ * Copyright © 2006-2011 Guillem Jover <guillem@debian.org>
+ * Copyright © 2009 Canonical Ltd.
+ * Copyright © 2011 Linaro Limited
+ * Copyright © 2011 Raphaël Hertzog <hertzog@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -29,8 +33,10 @@
 #include <dpkg/i18n.h>
 #include <dpkg/dpkg.h>
 #include <dpkg/dpkg-db.h>
+#include <dpkg/arch.h>
 #include <dpkg/path.h>
 #include <dpkg/parsedump.h>
+#include <dpkg/pkg-spec.h>
 #include <dpkg/triglib.h>
 
 static int
@@ -82,11 +88,11 @@ f_name(struct pkginfo *pigp, struct pkgbin *pifp,
 {
   const char *e;
 
-  e = pkg_name_is_illegal(value, NULL);
+  e = pkg_name_is_illegal(value);
   if (e != NULL)
     parse_error(ps, _("invalid package name (%.250s)"), e);
-  /* We use the new name, as pkg_db_find() may have done a tolower for us. */
-  pigp->name = pkg_db_find(value)->name;
+  /* We use the new name, as pkg_db_find_set() may have done a tolower for us. */
+  pigp->set->name = pkg_db_find_set(value)->name;
 }
 
 void
@@ -156,6 +162,32 @@ f_boolean(struct pkginfo *pigp, struct pkgbin *pifp,
   PKGPFIELD(pifp, fip->integer, bool) = boolean;
 }
 
+void
+f_multiarch(struct pkginfo *pigp, struct pkgbin *pifp,
+            struct parsedb_state *ps,
+            const char *value, const struct fieldinfo *fip)
+{
+  int multiarch;
+
+  if (!*value)
+    return;
+
+  multiarch = parse_nv_last(ps, _("foreign/allowed/same/no in quadstate field"),
+                            multiarchinfos, value);
+  PKGPFIELD(pifp, fip->integer, int) = multiarch;
+}
+
+void
+f_architecture(struct pkginfo *pigp, struct pkgbin *pifp,
+               struct parsedb_state *ps,
+               const char *value, const struct fieldinfo *fip)
+{
+  pifp->arch = dpkg_arch_find(value);
+  if (pifp->arch->type == arch_illegal)
+    parse_warn(ps, _("'%s' is not a valid architecture name: %s"),
+               value, dpkg_arch_name_is_illegal(value));
+}
+
 void
 f_section(struct pkginfo *pigp, struct pkgbin *pifp,
           struct parsedb_state *ps,
@@ -275,7 +307,8 @@ f_conffiles(struct pkginfo *pigp, struct pkgbin *pifp,
   static const char obsolete_str[]= "obsolete";
   struct conffile **lastp, *newlink;
   const char *endent, *endfn, *hashstart;
-  int c, namelen, hashlen, obsolete;
+  int c, namelen, hashlen;
+  bool obsolete;
   char *newptr;
 
   lastp= &pifp->conffiles;
@@ -353,25 +386,27 @@ f_dependency(struct pkginfo *pigp, struct pkgbin *pifp,
     for (;;) {
       depnamestart= p;
       /* Skip over package name characters. */
-      while (*p && !isspace(*p) && *p != '(' && *p != ',' && *p != '|') {
-	p++;
-      }
+      while (*p && !isspace(*p) && *p != ':' && *p != '(' && *p != ',' &&
+             *p != '|')
+        p++;
       depnamelength= p - depnamestart ;
-      varbuf_reset(&depname);
-      varbuf_add_buf(&depname, depnamestart, depnamelength);
-      varbuf_end_str(&depname);
-      if (!depname.buf[0])
+      if (depnamelength == 0)
         parse_error(ps,
                     _("`%s' field, missing package name, or garbage where "
                       "package name expected"), fip->name);
-      emsg = pkg_name_is_illegal(depname.buf, NULL);
+
+      varbuf_reset(&depname);
+      varbuf_add_buf(&depname, depnamestart, depnamelength);
+      varbuf_end_str(&depname);
+
+      emsg = pkg_name_is_illegal(depname.buf);
       if (emsg)
         parse_error(ps,
                     _("`%s' field, invalid package name `%.255s': %s"),
                     fip->name, depname.buf, emsg);
       dop= nfmalloc(sizeof(struct deppossi));
       dop->up= dyp;
-      dop->ed = pkg_db_find(depname.buf);
+      dop->ed = pkg_db_find_set(depname.buf);
       dop->next= NULL; *ldopp= dop; ldopp= &dop->next;
 
       /* Don't link this (which is after all only ‘newpig’ from
@@ -385,6 +420,48 @@ f_dependency(struct pkginfo *pigp, struct pkgbin *pifp,
 
       dop->cyclebreak = false;
 
+      /* See if we have an architecture qualifier. */
+      if (*p == ':') {
+        static struct varbuf arch;
+        const char *archstart;
+        int archlength;
+
+        archstart = ++p;
+        while (*p && !isspace(*p) && *p != '(' && *p != ',' && *p != '|')
+          p++;
+        archlength = p - archstart;
+        if (archlength == 0)
+          parse_error(ps, _("'%s' field, missing architecture name, or garbage "
+                            "where architecture name expected"), fip->name);
+
+        varbuf_reset(&arch);
+        varbuf_add_buf(&arch, archstart, archlength);
+        varbuf_end_str(&arch);
+
+        dop->arch_is_implicit = false;
+        dop->arch = dpkg_arch_find(arch.buf);
+
+        if (dop->arch->type == arch_illegal)
+          emsg = dpkg_arch_name_is_illegal(arch.buf);
+        else if (dop->arch->type != arch_wildcard)
+          emsg = _("a value different from 'any' is currently not allowed");
+        if (emsg)
+          parse_error(ps, _("'%s' field, reference to '%.255s': "
+                            "invalid architecture name '%.255s': %s"),
+                      fip->name, depname.buf, arch.buf, emsg);
+      } else if (fip->integer == dep_conflicts || fip->integer == dep_breaks ||
+                 fip->integer == dep_replaces) {
+        /* Conflics/Breaks/Replaces get an implicit "any" arch qualifier. */
+        dop->arch_is_implicit = true;
+        dop->arch = dpkg_arch_get(arch_wildcard);
+      } else {
+        /* Otherwise use the pkgbin architecture, which will be assigned to
+         * later on by parse.c, once we can guarantee we have parsed it from
+         * the control stanza. */
+        dop->arch_is_implicit = true;
+        dop->arch = NULL;
+      }
+
       /* Skip whitespace after package name. */
       while (isspace(*p)) p++;
 
@@ -553,7 +630,7 @@ f_trigaw(struct pkginfo *aw, struct pkgbin *pifp,
          struct parsedb_state *ps,
          const char *value, const struct fieldinfo *fip)
 {
-  const char *word, *emsg;
+  const char *word;
   struct pkginfo *pend;
 
   if (ps->flags & pdb_rejectstatus)
@@ -562,12 +639,13 @@ f_trigaw(struct pkginfo *aw, struct pkgbin *pifp,
                   "this context"));
 
   while ((word = scan_word(&value))) {
-    emsg = pkg_name_is_illegal(word, NULL);
-    if (emsg)
+    struct dpkg_error err;
+
+    pend = pkg_spec_parse_pkg(word, &err);
+    if (pend == NULL)
       parse_error(ps,
                   _("illegal package name in awaited trigger `%.255s': %s"),
-                  word, emsg);
-    pend = pkg_db_find(word);
+                  word, err.str);
 
     if (!trig_note_aw(pend, aw))
       parse_error(ps,
