@@ -3,6 +3,7 @@
  * parsehelp.c - helpful routines for parsing and writing
  *
  * Copyright © 1995 Ian Jackson <ian@chiark.greenend.org.uk>
+ * Copyright © 2006-2011 Guillem Jover <guillem@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -41,9 +42,10 @@ parse_error_msg(struct parsedb_state *ps, const char *fmt)
 
   str_escape_fmt(filename, ps->filename, sizeof(filename));
 
-  if (ps->pkg && ps->pkg->name)
+  if (ps->pkg && ps->pkg->set->name)
     sprintf(msg, _("parsing file '%.255s' near line %d package '%.255s':\n"
-                   " %.255s"), filename, ps->lno, ps->pkg->name, fmt);
+                   " %.255s"), filename, ps->lno,
+                   pkgbin_name(ps->pkg, ps->pkgbin, pnaw_nonambig), fmt);
   else
     sprintf(msg, _("parsing file '%.255s' near line %d:\n"
                    " %.255s"), filename, ps->lno, fmt);
@@ -76,6 +78,14 @@ const struct namevalue booleaninfos[] = {
   { .name = NULL }
 };
 
+const struct namevalue multiarchinfos[] = {
+  NAMEVALUE_DEF("no",      multiarch_no),
+  NAMEVALUE_DEF("same",    multiarch_same),
+  NAMEVALUE_DEF("allowed", multiarch_allowed),
+  NAMEVALUE_DEF("foreign", multiarch_foreign),
+  { .name = NULL }
+};
+
 const struct namevalue priorityinfos[] = {
   NAMEVALUE_DEF("required",                      pri_required),
   NAMEVALUE_DEF("important",                     pri_important),
@@ -115,7 +125,7 @@ const struct namevalue wantinfos[] = {
 };
 
 const char *
-pkg_name_is_illegal(const char *p, const char **ep)
+pkg_name_is_illegal(const char *p)
 {
   /* FIXME: _ is deprecated, remove sometime. */
   static const char alsoallowed[] = "-+._";
@@ -128,10 +138,7 @@ pkg_name_is_illegal(const char *p, const char **ep)
   while ((c = *p++) != '\0')
     if (!isalnum(c) && !strchr(alsoallowed,c)) break;
   if (!c) return NULL;
-  if (isspace(c) && ep) {
-    while (isspace(*p)) p++;
-    *ep= p; return NULL;
-  }
+
   snprintf(buf, sizeof(buf), _(
 	   "character `%c' not allowed (only letters, digits and characters `%s')"),
 	   c, alsoallowed);
