@@ -42,11 +42,11 @@ pkg_array_init_from_db(struct pkg_array *a)
 	struct pkginfo *pkg;
 	int i;
 
-	a->n_pkgs = pkg_db_count();
+	a->n_pkgs = pkg_db_count_pkg();
 	a->pkgs = m_malloc(sizeof(a->pkgs[0]) * a->n_pkgs);
 
 	it = pkg_db_iter_new();
-	for (i = 0; (pkg = pkg_db_iter_next(it)); i++)
+	for (i = 0; (pkg = pkg_db_iter_next_pkg(it)); i++)
 		a->pkgs[i] = pkg;
 	pkg_db_iter_free(it);
 
