@@ -3,6 +3,7 @@
  * pkg-format.c - customizable package formatting
  *
  * Copyright © 2001 Wichert Akkerman <wakkerma@debian.org>
+ * Copyright © 2008-2012 Guillem Jover <guillem@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -29,6 +30,7 @@
 #include <dpkg/dpkg.h>
 #include <dpkg/dpkg-db.h>
 #include <dpkg/parsedump.h>
+#include <dpkg/pkg-show.h>
 #include <dpkg/pkg-format.h>
 
 enum pkg_format_type {
@@ -200,6 +202,107 @@ pkg_format_parse(const char *fmt)
 	return head;
 }
 
+static void
+virt_package(struct varbuf *vb,
+             const struct pkginfo *pkg, const struct pkgbin *pkgbin,
+             enum fwriteflags flags, const struct fieldinfo *fip)
+{
+	varbuf_add_pkgbin_name(vb, pkg, pkgbin, pnaw_nonambig);
+}
+
+static void
+virt_status_abbrev(struct varbuf *vb,
+                   const struct pkginfo *pkg, const struct pkgbin *pkgbin,
+                   enum fwriteflags flags, const struct fieldinfo *fip)
+{
+	if (pkgbin != &pkg->installed)
+		return;
+
+	varbuf_add_char(vb, pkg_abbrev_want(pkg));
+	varbuf_add_char(vb, pkg_abbrev_status(pkg));
+	varbuf_add_char(vb, pkg_abbrev_eflag(pkg));
+}
+
+static void
+virt_summary(struct varbuf *vb,
+             const struct pkginfo *pkg, const struct pkgbin *pkgbin,
+             enum fwriteflags flags, const struct fieldinfo *fip)
+{
+	const char *desc;
+	int len;
+
+	desc = pkg_summary(pkg, pkgbin, &len);
+
+	varbuf_add_buf(vb, desc, len);
+}
+
+static void
+virt_source_package(struct varbuf *vb,
+                    const struct pkginfo *pkg, const struct pkgbin *pkgbin,
+                    enum fwriteflags flags, const struct fieldinfo *fip)
+{
+	const char *name;
+	size_t len;
+
+	name = pkgbin->source;
+	if (name == NULL)
+		name = pkg->set->name;
+
+	len = strcspn(name, " ");
+	if (len == 0)
+		len = strlen(name);
+
+	varbuf_add_buf(vb, name, len);
+}
+
+static void
+virt_source_version(struct varbuf *vb,
+                    const struct pkginfo *pkg, const struct pkgbin *pkgbin,
+                    enum fwriteflags flags, const struct fieldinfo *fip)
+{
+	const char *version;
+	size_t len;
+
+	if (pkgbin->source)
+		version = strchr(pkgbin->source, '(');
+	else
+		version = NULL;
+
+	if (version == NULL) {
+		varbufversion(vb, &pkgbin->version, vdew_nonambig);
+	} else {
+		version++;
+
+		len = strcspn(version, ")");
+		if (len == 0)
+			len = strlen(version);
+
+		varbuf_add_buf(vb, version, len);
+	}
+}
+
+const struct fieldinfo virtinfos[] = {
+	{ "binary:Package", NULL, virt_package },
+	{ "binary:Summary", NULL, virt_summary },
+	{ "db:Status-Abbrev", NULL, virt_status_abbrev },
+	{ "source:Package", NULL, virt_source_package },
+	{ "source:Version", NULL, virt_source_version },
+	{ NULL },
+};
+
+static const struct fieldinfo *
+find_field_info(const struct fieldinfo *fields_head,
+                const struct pkg_format_node *fmt_node)
+{
+	const struct fieldinfo *fip;
+
+	for (fip = fields_head; fip->name; fip++)
+		if (strcasecmp(fmt_node->data, fip->name) == 0)
+			break;
+
+	return fip;
+}
+
 void
 pkg_format_show(const struct pkg_format_node *head,
                 struct pkginfo *pkg, struct pkgbin *pif)
@@ -224,18 +327,19 @@ pkg_format_show(const struct pkg_format_node *head,
 		} else if (head->type == field) {
 			const struct fieldinfo *fip;
 
-			for (fip = fieldinfos; fip->name; fip++)
-				if (strcasecmp(head->data, fip->name) == 0) {
-					fip->wcall(&wb, pkg, pif, 0, fip);
+			fip = find_field_info(fieldinfos, head);
 
-					varbuf_end_str(&wb);
-					varbuf_printf(&fb, fmt, wb.buf);
-					varbuf_reset(&wb);
-					ok = true;
-					break;
-				}
+			if (fip->name == NULL)
+				fip = find_field_info(virtinfos, head);
 
-			if (!fip->name) {
+			if (fip->name) {
+				fip->wcall(&wb, pkg, pif, 0, fip);
+
+				varbuf_end_str(&wb);
+				varbuf_printf(&fb, fmt, wb.buf);
+				varbuf_reset(&wb);
+				ok = true;
+			} else {
 				const struct arbitraryfield *afp;
 
 				for (afp = pif->arbs; afp; afp = afp->next)
