@@ -3,7 +3,7 @@
  * pkg.c - primitives for pkg handling
  *
  * Copyright © 1995, 1996 Ian Jackson <ian@chiark.greenend.org.uk>
- * Copyright © 2009 Guillem Jover <guillem@debian.org>
+ * Copyright © 2009-2012 Guillem Jover <guillem@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,24 +22,88 @@
 #include <config.h>
 #include <compat.h>
 
+#include <assert.h>
 #include <string.h>
 
 #include <dpkg/dpkg-db.h>
 #include <dpkg/pkg.h>
 
+/**
+ * Set the package installation status.
+ */
+void
+pkg_set_status(struct pkginfo *pkg, enum pkgstatus status)
+{
+	if (pkg->status == status)
+		return;
+	else if (pkg->status == stat_notinstalled)
+		pkg->set->installed_instances++;
+	else if (status == stat_notinstalled)
+		pkg->set->installed_instances--;
+
+	assert(pkg->set->installed_instances >= 0);
+
+	pkg->status = status;
+}
+
+/**
+ * Set the specified flags to 1 in the package error flags.
+ */
+void
+pkg_set_eflags(struct pkginfo *pkg, enum pkgeflag eflag)
+{
+	pkg->eflag |= eflag;
+}
+
+/**
+ * Clear the specified flags to 0 in the package error flags.
+ */
+void
+pkg_clear_eflags(struct pkginfo *pkg, enum pkgeflag eflag)
+{
+	pkg->eflag &= ~eflag;
+}
+
+/**
+ * Reset the package error flags to 0.
+ */
+void
+pkg_reset_eflags(struct pkginfo *pkg)
+{
+	pkg->eflag = eflag_ok;
+}
+
+/**
+ * Copy the package error flags to another package.
+ */
+void
+pkg_copy_eflags(struct pkginfo *pkg_dst, struct pkginfo *pkg_src)
+{
+	pkg_dst->eflag = pkg_src->eflag;
+}
+
+/**
+ * Set the package selection status.
+ */
+void
+pkg_set_want(struct pkginfo *pkg, enum pkgwant want)
+{
+	pkg->want = want;
+}
+
 void
 pkgbin_blank(struct pkgbin *pkgbin)
 {
 	pkgbin->essential = false;
+	pkgbin->multiarch = multiarch_no;
 	pkgbin->depends = NULL;
-	pkgbin->depended = NULL;
+	pkgbin->pkgname_archqual = NULL;
 	pkgbin->description = NULL;
 	pkgbin->maintainer = NULL;
 	pkgbin->source = NULL;
 	pkgbin->installedsize = NULL;
 	pkgbin->bugs = NULL;
 	pkgbin->origin = NULL;
-	pkgbin->arch = NULL;
 	blankversion(&pkgbin->version);
 	pkgbin->conffiles = NULL;
 	pkgbin->arbs = NULL;
@@ -48,7 +112,6 @@ pkgbin_blank(struct pkgbin *pkgbin)
 void
 pkg_blank(struct pkginfo *pkg)
 {
-	pkg->name = NULL;
 	pkg->status = stat_notinstalled;
 	pkg->eflag = eflag_ok;
 	pkg->want = want_unknown;
@@ -64,6 +127,51 @@ pkg_blank(struct pkginfo *pkg)
 	pkg->trigpend_head = NULL;
 	pkgbin_blank(&pkg->installed);
 	pkgbin_blank(&pkg->available);
+
+	/* The architectures are reset here (instead of in pkgbin_blank),
+	 * because they are part of the package specification, and needed
+	 * for selections. */
+	pkg->installed.arch = dpkg_arch_get(arch_none);
+	pkg->available.arch = dpkg_arch_get(arch_none);
+}
+
+void
+pkgset_blank(struct pkgset *set)
+{
+	set->name = NULL;
+	set->depended.available = NULL;
+	set->depended.installed = NULL;
+	pkg_blank(&set->pkg);
+	set->installed_instances = 0;
+	set->pkg.set = set;
+	set->pkg.arch_next = NULL;
+}
+
+/**
+ * Link a pkginfo instance into a package set.
+ *
+ * @param set The package set to use.
+ * @param pkg The package to link into the set.
+ */
+void
+pkgset_link_pkg(struct pkgset *set, struct pkginfo *pkg)
+{
+	pkg->set = set;
+	pkg->arch_next = set->pkg.arch_next;
+	set->pkg.arch_next = pkg;
+}
+
+/**
+ * Get the number of installed package instances in a package set.
+ *
+ * @param set The package set to use.
+ *
+ * @return The count of installed packages.
+ */
+int
+pkgset_installed_instances(struct pkgset *set)
+{
+	return set->installed_instances;
 }
 
 static int
@@ -103,23 +211,3 @@ pkg_is_informative(struct pkginfo *pkg, struct pkgbin *pkgbin)
 
 	return false;
 }
-
-/**
- * Compare a package to be sorted by name.
- *
- * @param a A pointer of a pointer to a struct pkginfo.
- * @param b A pointer of a pointer to a struct pkginfo.
- *
- * @return An integer with the result of the comparison.
- * @retval -1 a is earlier than b.
- * @retval 0 a is equal to b.
- * @retval 1 a is later than b.
- */
-int
-pkg_sorter_by_name(const void *a, const void *b)
-{
-	const struct pkginfo *pa = *(const struct pkginfo **)a;
-	const struct pkginfo *pb = *(const struct pkginfo **)b;
-
-	return strcmp(pa->name, pb->name);
-}
