@@ -4,7 +4,7 @@
  *
  * Copyright © 2007 Canonical Ltd
  * Written by Ian Jackson <ian@chiark.greenend.org.uk>
- * Copyright © 2008-2011 Guillem Jover <guillem@debian.org>
+ * Copyright © 2008-2012 Guillem Jover <guillem@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -34,14 +34,16 @@
 #include <dpkg/i18n.h>
 #include <dpkg/dpkg.h>
 #include <dpkg/dpkg-db.h>
+#include <dpkg/pkg.h>
 #include <dpkg/dlist.h>
 #include <dpkg/dir.h>
+#include <dpkg/pkg-spec.h>
 #include <dpkg/trigdeferred.h>
 #include <dpkg/triglib.h>
 
 /*========== Recording triggers. ==========*/
 
-static char *triggersdir, *triggersfilefile, *triggersnewfilefile;
+static char *triggersdir, *triggersfilefile;
 
 static char *
 trig_get_filename(const char *dir, const char *filename)
@@ -94,7 +96,7 @@ trig_record_activation(struct pkginfo *pend, struct pkginfo *aw, const char *tri
 	if (aw && pend->status > stat_configfiles)
 		if (trig_note_aw(pend, aw)) {
 			if (aw->status > stat_triggersawaited)
-				aw->status = stat_triggersawaited;
+				pkg_set_status(aw, stat_triggersawaited);
 			modstatdb_note_ifwrite(aw);
 		}
 }
@@ -115,8 +117,10 @@ trig_clear_awaiters(struct pkginfo *notpend)
 			continue;
 		LIST_UNLINK_PART(aw->trigaw, ta, sameaw.);
 		if (!aw->trigaw.head && aw->status == stat_triggersawaited) {
-			aw->status = aw->trigpend_head ? stat_triggerspending :
-			             stat_installed;
+			if (aw->trigpend_head)
+				pkg_set_status(aw, stat_triggerspending);
+			else
+				pkg_set_status(aw, stat_installed);
 			modstatdb_note(aw);
 		}
 	}
@@ -150,8 +154,9 @@ struct trigkindinfo {
 	/* Rest are for everyone: */
 	void (*activate_awaiter)(struct pkginfo *pkg /* may be NULL */);
 	void (*activate_done)(void);
-	void (*interest_change)(const char *name, struct pkginfo *pkg, int signum,
-	                        enum trig_options opts);
+	void (*interest_change)(const char *name, struct pkginfo *pkg,
+	                        struct pkgbin *pkgbin,
+	                        int signum, enum trig_options opts);
 };
 
 static const struct trigkindinfo tki_explicit, tki_file, tki_unknown;
@@ -176,7 +181,7 @@ trig_classify_byname(const char *name)
 		return &tki_file;
 	}
 
-	if (!pkg_name_is_illegal(name, NULL) && !strchr(name, '_'))
+	if (!pkg_name_is_illegal(name) && !strchr(name, '_'))
 		return &tki_explicit;
 
 invalid:
@@ -216,12 +221,13 @@ trk_unknown_activate_done(void)
 }
 
 static void DPKG_ATTR_NORET
-trk_unknown_interest_change(const char *trig, struct pkginfo *pkg, int signum,
+trk_unknown_interest_change(const char *trig, struct pkginfo *pkg,
+                            struct pkgbin *pkgbin, int signum,
                             enum trig_options opts)
 {
 	ohshit(_("invalid or unknown syntax in trigger name `%.250s'"
 	         " (in trigger interests for package `%.250s')"),
-	       trig, pkg->name);
+	       trig, pkgbin_name(pkg, pkgbin, pnaw_nonambig));
 }
 
 static const struct trigkindinfo tki_unknown = {
@@ -282,7 +288,6 @@ static void
 trk_explicit_activate_awaiter(struct pkginfo *aw)
 {
 	char buf[1024];
-	const char *emsg;
 	struct pkginfo *pend;
 
 	if (!trk_explicit_f)
@@ -293,6 +298,7 @@ trk_explicit_activate_awaiter(struct pkginfo *aw)
 		        trk_explicit_fn.buf);
 
 	while (trk_explicit_fgets(buf, sizeof(buf)) >= 0) {
+		struct dpkg_error err;
 		char *slash;
 		bool noawait = false;
 		slash = strchr(buf, '/');
@@ -300,93 +306,59 @@ trk_explicit_activate_awaiter(struct pkginfo *aw)
 			noawait = true;
 			*slash = '\0';
 		}
-		emsg = pkg_name_is_illegal(buf, NULL);
-		if (emsg)
+
+		pend = pkg_spec_parse_pkg(buf, &err);
+		if (pend == NULL)
 			ohshit(_("trigger interest file `%.250s' syntax error; "
 			         "illegal package name `%.250s': %.250s"),
-			       trk_explicit_fn.buf, buf, emsg);
-		pend = pkg_db_find(buf);
+			       trk_explicit_fn.buf, buf, err.str);
+
 		trig_record_activation(pend, noawait ? NULL : aw,
 		                       trk_explicit_trig);
 	}
 }
 
 static void
-trk_explicit_interest_flush(const char *newfilename, FILE *nf)
-{
-	if (ferror(nf))
-		ohshite(_("unable to write new trigger interest file `%.250s'"),
-		        newfilename);
-	if (fflush(nf))
-		ohshite(_("unable to flush new trigger interest file '%.250s'"),
-		        newfilename);
-	if (fsync(fileno(nf)))
-		ohshite(_("unable to sync new trigger interest file '%.250s'"),
-		        newfilename);
-}
-
-static void
-trk_explicit_interest_commit(const char *newfilename)
-{
-	if (rename(newfilename, trk_explicit_fn.buf))
-		ohshite(_("unable to install new trigger interest file `%.250s'"),
-		        trk_explicit_fn.buf);
-}
-
-static void
-trk_explicit_interest_remove(const char *newfilename)
-{
-	if (unlink(newfilename))
-		ohshite(_("cannot remove `%.250s'"), newfilename);
-	if (unlink(trk_explicit_fn.buf) && errno != ENOENT)
-		ohshite(_("cannot remove `%.250s'"), trk_explicit_fn.buf);
-}
-
-static void
-trk_explicit_interest_change(const char *trig,  struct pkginfo *pkg, int signum,
+trk_explicit_interest_change(const char *trig,  struct pkginfo *pkg,
+                             struct pkgbin *pkgbin, int signum,
                              enum trig_options opts)
 {
-	static struct varbuf newfn;
 	char buf[1024];
-	FILE *nf;
+	struct atomic_file *file;
 	bool empty = true;
 
 	trk_explicit_start(trig);
-	varbuf_reset(&newfn);
-	varbuf_printf(&newfn, "%s/%s.new", triggersdir, trig);
-
-	nf = fopen(newfn.buf, "w");
-	if (!nf)
-		ohshite(_("unable to create new trigger interest file `%.250s'"),
-		        newfn.buf);
-	push_cleanup(cu_closestream, ~ehflag_normaltidy, NULL, 0, 1, nf);
+	file = atomic_file_new(trk_explicit_fn.buf, 0);
+	atomic_file_open(file);
 
 	while (trk_explicit_f && trk_explicit_fgets(buf, sizeof(buf)) >= 0) {
-		int len = strlen(pkg->name);
-		if (strncmp(buf, pkg->name, len) == 0 &&
+		const char *pkgname = pkgbin_name(pkg, pkgbin, pnaw_nonambig);
+		int len = strlen(pkgname);
+
+		if (strncmp(buf, pkgname, len) == 0 &&
 		    (buf[len] == '\0' || buf[len] == '/'))
 			continue;
-		fprintf(nf, "%s\n", buf);
+		fprintf(file->fp, "%s\n", buf);
 		empty = false;
 	}
 	if (signum > 0) {
-		fprintf(nf, "%s%s\n", pkg->name,
+		fprintf(file->fp, "%s%s\n",
+		        pkgbin_name(pkg, pkgbin, pnaw_nonambig),
 		        (opts == trig_noawait) ? "/noawait" : "");
 		empty = false;
 	}
 
 	if (!empty)
-		trk_explicit_interest_flush(newfn.buf, nf);
+		atomic_file_sync(file);
 
-	pop_cleanup(ehflag_normaltidy);
-	if (fclose(nf))
-		ohshite(_("unable to close new trigger interest file `%.250s'"),
-		        newfn.buf);
+	atomic_file_close(file);
 
 	if (empty)
-		trk_explicit_interest_remove(newfn.buf);
+		atomic_file_remove(file);
 	else
-		trk_explicit_interest_commit(newfn.buf);
+		atomic_file_commit(file);
+
+	atomic_file_free(file);
 
 	dir_sync_path(triggersdir);
 }
@@ -418,7 +390,8 @@ static int filetriggers_edited = -1;
  * but die if already present.
  */
 static void
-trk_file_interest_change(const char *trig, struct pkginfo *pkg, int signum,
+trk_file_interest_change(const char *trig, struct pkginfo *pkg,
+                         struct pkgbin *pkgbin, int signum,
                          enum trig_options opts)
 {
 	struct filenamenode *fnn;
@@ -442,6 +415,7 @@ trk_file_interest_change(const char *trig, struct pkginfo *pkg, int signum,
 
 	tfi = nfmalloc(sizeof(*tfi));
 	tfi->pkg = pkg;
+	tfi->pkgbin = pkgbin;
 	tfi->fnn = fnn;
 	tfi->options = opts;
 	tfi->samefile_next = *trigh.namenode_interested(fnn);
@@ -454,7 +428,8 @@ found:
 	tfi->options = opts;
 	if (signum > 1)
 		ohshit(_("duplicate file trigger interest for filename `%.250s' "
-		         "and package `%.250s'"), trig, pkg->name);
+		         "and package `%.250s'"), trig,
+		       pkgbin_name(pkg, pkgbin, pnaw_nonambig));
 	if (signum > 0)
 		return;
 
@@ -476,36 +451,20 @@ static void
 trig_file_interests_update(void)
 {
 	struct trigfileint *tfi;
-	FILE *nf;
+	struct atomic_file *file;
 
-	nf = fopen(triggersnewfilefile, "w");
-	if (!nf)
-		ohshite(_("unable to create new file triggers file `%.250s'"),
-		        triggersnewfilefile);
-	push_cleanup(cu_closestream, ~ehflag_normaltidy, NULL, 0, 1, nf);
+	file = atomic_file_new(triggersfilefile, 0);
+	atomic_file_open(file);
 
 	for (tfi = filetriggers.head; tfi; tfi = tfi->inoverall.next)
-		fprintf(nf, "%s %s%s\n", trigh.namenode_name(tfi->fnn),
-		        tfi->pkg->name,
+		fprintf(file->fp, "%s %s%s\n", trigh.namenode_name(tfi->fnn),
+		        pkgbin_name(tfi->pkg, tfi->pkgbin, pnaw_nonambig),
 		        (tfi->options == trig_noawait) ? "/noawait" : "");
 
-	if (ferror(nf))
-		ohshite(_("unable to write new file triggers file `%.250s'"),
-		        triggersnewfilefile);
-	if (fflush(nf))
-		ohshite(_("unable to flush new file triggers file '%.250s'"),
-		        triggersnewfilefile);
-	if (fsync(fileno(nf)))
-		ohshite(_("unable to sync new file triggers file '%.250s'"),
-		        triggersnewfilefile);
-	pop_cleanup(ehflag_normaltidy);
-	if (fclose(nf))
-		ohshite(_("unable to close new file triggers file `%.250s'"),
-		        triggersnewfilefile);
-
-	if (rename(triggersnewfilefile, triggersfilefile))
-		ohshite(_("unable to install new file triggers file as `%.250s'"),
-		        triggersfilefile);
+	atomic_file_sync(file);
+	atomic_file_close(file);
+	atomic_file_commit(file);
+	atomic_file_free(file);
 }
 
 void
@@ -530,7 +489,7 @@ trig_file_interests_ensure(void)
 	FILE *f;
 	char linebuf[1024], *space;
 	struct pkginfo *pkg;
-	const char *emsg;
+	struct pkgbin *pkgbin;
 
 	if (filetriggers_edited >= 0)
 		return;
@@ -545,6 +504,7 @@ trig_file_interests_ensure(void)
 
 	push_cleanup(cu_closestream, ~0, NULL, 0, 1, f);
 	while (fgets_checked(linebuf, sizeof(linebuf), f, triggersfilefile) >= 0) {
+		struct dpkg_error err;
 		char *slash;
 		enum trig_options trig_opts = trig_await;
 		space = strchr(linebuf, ' ');
@@ -558,13 +518,15 @@ trig_file_interests_ensure(void)
 			trig_opts = trig_noawait;
 			*slash = '\0';
 		}
-		emsg = pkg_name_is_illegal(space, NULL);
-		if (emsg)
+
+		pkg = pkg_spec_parse_pkg(space, &err);
+		if (pkg == NULL)
 			ohshit(_("file triggers record mentions illegal "
 			         "package name `%.250s' (for interest in file "
-				 "`%.250s'): %.250s"), space, linebuf, emsg);
-		pkg = pkg_db_find(space);
-		trk_file_interest_change(linebuf, pkg, +2, trig_opts);
+			         "`%.250s'): %.250s"), space, linebuf, err.str);
+		pkgbin = &pkg->installed;
+
+		trk_file_interest_change(linebuf, pkg, pkgbin, +2, trig_opts);
 	}
 	pop_cleanup(ehflag_normaltidy);
 ok:
@@ -622,32 +584,35 @@ static const struct trigkindinfo tki_file = {
 /*---------- Trigger control info file. ----------*/
 
 static void
-trig_cicb_interest_change(const char *trig, struct pkginfo *pkg, int signum,
+trig_cicb_interest_change(const char *trig, struct pkginfo *pkg,
+                          struct pkgbin *pkgbin, int signum,
                           enum trig_options opts)
 {
 	const struct trigkindinfo *tki = trig_classify_byname(trig);
 
 	assert(filetriggers_edited >= 0);
-	tki->interest_change(trig, pkg, signum, opts);
+	tki->interest_change(trig, pkg, pkgbin, signum, opts);
 }
 
 void
-trig_cicb_interest_delete(const char *trig, void *user, enum trig_options opts)
+trig_cicb_interest_delete(const char *trig, struct pkginfo *pkg,
+                          struct pkgbin *pkgbin, enum trig_options opts)
 {
-	trig_cicb_interest_change(trig, user, -1, opts);
+	trig_cicb_interest_change(trig, pkg, pkgbin, -1, opts);
 }
 
 void
-trig_cicb_interest_add(const char *trig, void *user, enum trig_options opts)
+trig_cicb_interest_add(const char *trig, struct pkginfo *pkg,
+                       struct pkgbin *pkgbin, enum trig_options opts)
 {
-	trig_cicb_interest_change(trig, user, +1, opts);
+	trig_cicb_interest_change(trig, pkg, pkgbin, +1, opts);
 }
 
 void
-trig_cicb_statuschange_activate(const char *trig, void *user,
-                                enum trig_options opts)
+trig_cicb_statuschange_activate(const char *trig, struct pkginfo *pkg,
+                                struct pkgbin *pkgbin, enum trig_options opts)
 {
-	struct pkginfo *aw = user;
+	struct pkginfo *aw = pkg;
 
 	trig_activate_start(trig);
 	dtki->activate_awaiter((opts == trig_noawait) ? NULL : aw);
@@ -656,7 +621,8 @@ trig_cicb_statuschange_activate(const char *trig, void *user,
 
 static void
 parse_ci_call(const char *file, const char *cmd, trig_parse_cicb *cb,
-              const char *trig, void *user, enum trig_options opts)
+              const char *trig, struct pkginfo *pkg, struct pkgbin *pkgbin,
+              enum trig_options opts)
 {
 	const char *emsg;
 
@@ -666,12 +632,13 @@ parse_ci_call(const char *file, const char *cmd, trig_parse_cicb *cb,
 		         "syntax in trigger name `%.250s': %.250s"),
 		       file, trig, emsg);
 	if (cb)
-		cb(trig, user, opts);
+		cb(trig, pkg, pkgbin, opts);
 }
 
 void
 trig_parse_ci(const char *file, trig_parse_cicb *interest,
-              trig_parse_cicb *activate, void *user)
+              trig_parse_cicb *activate, struct pkginfo *pkg,
+              struct pkgbin *pkgbin)
 {
 	FILE *f;
 	char linebuf[MAXTRIGDIRECTIVE], *cmd, *spc, *eol;
@@ -701,13 +668,13 @@ trig_parse_ci(const char *file, trig_parse_cicb *interest,
 		while (cisspace(*spc))
 			spc++;
 		if (!strcmp(cmd, "interest")) {
-			parse_ci_call(file, cmd, interest, spc, user, trig_await);
+			parse_ci_call(file, cmd, interest, spc, pkg, pkgbin, trig_await);
 		} else if (!strcmp(cmd, "interest-noawait")) {
-			parse_ci_call(file, cmd, interest, spc, user, trig_noawait);
+			parse_ci_call(file, cmd, interest, spc, pkg, pkgbin, trig_noawait);
 		} else if (!strcmp(cmd, "activate")) {
-			parse_ci_call(file, cmd, activate, spc, user, trig_await);
+			parse_ci_call(file, cmd, activate, spc, pkg, pkgbin, trig_await);
 		} else if (!strcmp(cmd, "activate-noawait")) {
-			parse_ci_call(file, cmd, activate, spc, user, trig_noawait);
+			parse_ci_call(file, cmd, activate, spc, pkg, pkgbin, trig_noawait);
 		} else {
 			ohshit(_("triggers ci file contains unknown directive `%.250s'"),
 			       cmd);
@@ -727,7 +694,12 @@ tdm_incorp_trig_begin(const char *trig)
 static void
 tdm_incorp_package(const char *awname)
 {
-	struct pkginfo *aw = strcmp(awname, "-") ? pkg_db_find(awname) : NULL;
+	struct pkginfo *aw;
+
+	if (strcmp(awname, "-") == 0)
+		aw = NULL;
+	else
+		aw = pkg_spec_parse_pkg(awname, NULL);
 
 	dtki->activate_awaiter(aw);
 }
@@ -756,9 +728,6 @@ trig_incorporate(enum modstatdb_rw cstatus)
 	free(triggersfilefile);
 	triggersfilefile = trig_get_filename(triggersdir, "File");
 
-	free(triggersnewfilefile);
-	triggersnewfilefile = trig_get_filename(triggersdir, "File.new");
-
 	trigdef_set_methods(&tdm_incorp);
 	trig_file_interests_ensure();
 
