@@ -52,8 +52,6 @@ BEGIN {
 sub version {
     printf _g("Debian %s version %s.\n"), $progname, $version;
 
-    printf "\n" . _g("Copyright (C) 2009-2010 Raphael Hertzog.");
-
     printf "\n" . _g(
 "This is free software; see the GNU General Public License version 2 or
 later for copying conditions. There is NO warranty.
@@ -179,7 +177,9 @@ sub compare_versions {
 	$a =~ s/~[^~]*$//;
 	$b =~ s/~[^~]*$//;
     }
-    return version_compare($a, $b);
+    $a = Dpkg::Version->new($a);
+    $b = Dpkg::Version->new($b);
+    return $a <=> $b;
 }
 
 # Merge changelog entries smartly by merging individually the different
