@@ -4,7 +4,9 @@
  *
  * Copyright © 1994,1995 Ian Jackson <ian@chiark.greenend.org.uk>
  * Copyright © 2000 Wichert Akkerman <wakkerma@debian.org>
- * Copyright © 2007-2011 Guillem Jover <guillem@debian.org>
+ * Copyright © 2007-2012 Guillem Jover <guillem@debian.org>
+ * Copyright © 2011 Linaro Limited
+ * Copyright © 2011 Raphaël Hertzog <hertzog@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -44,6 +46,7 @@
 #include <dpkg/i18n.h>
 #include <dpkg/dpkg.h>
 #include <dpkg/dpkg-db.h>
+#include <dpkg/pkg.h>
 #include <dpkg/path.h>
 #include <dpkg/fdio.h>
 #include <dpkg/buffer.h>
@@ -62,6 +65,7 @@
 #include "main.h"
 #include "archives.h"
 #include "filters.h"
+#include "infodb.h"
 
 static inline void
 fd_writeback_init(int fd)
@@ -114,18 +118,19 @@ filesavespackage(struct fileinlist *file,
                  struct pkginfo *pkgbeinginstalled)
 {
   struct filepackages_iterator *iter;
-  struct pkginfo *divpkg, *thirdpkg;
+  struct pkgset *divpkgset;
+  struct pkginfo *thirdpkg;
 
   debug(dbg_eachfiledetail,"filesavespackage file `%s' package %s",
-        file->namenode->name,pkgtobesaved->name);
+        file->namenode->name, pkg_name(pkgtobesaved, pnaw_always));
 
   /* If the file is a contended one and it's overridden by either
    * the package we're considering disappearing or the package
    * we're installing then they're not actually the same file, so
    * we can't disappear the package - it is saved by this file. */
   if (file->namenode->divert && file->namenode->divert->useinstead) {
-    divpkg= file->namenode->divert->pkg;
-    if (divpkg == pkgtobesaved || divpkg == pkgbeinginstalled) {
+    divpkgset = file->namenode->divert->pkgset;
+    if (divpkgset == pkgtobesaved->set || divpkgset == pkgbeinginstalled->set) {
       debug(dbg_eachfiledetail,"filesavespackage ... diverted -- save!");
       return true;
     }
@@ -140,13 +145,20 @@ filesavespackage(struct fileinlist *file,
   iter = filepackages_iter_new(file->namenode);
   while ((thirdpkg = filepackages_iter_next(iter))) {
     debug(dbg_eachfiledetail, "filesavespackage ... also in %s",
-          thirdpkg->name);
+          pkg_name(thirdpkg, pnaw_always));
 
     /* Is this not the package being installed or the one being
      * checked for disappearance? */
     if (thirdpkg == pkgbeinginstalled || thirdpkg == pkgtobesaved)
       continue;
 
+    /* A Multi-Arch: same package can share files and their presence in a
+     * third package of the same set is not a sign that we can get rid of
+     * it. */
+    if (pkgtobesaved->installed.multiarch == multiarch_same &&
+        thirdpkg->set == pkgtobesaved->set)
+      continue;
+
     /* If !fileslistvalid then we've already disappeared this one, so
      * we shouldn't try to make it take over this shared directory. */
     debug(dbg_eachfiledetail,"filesavespackage ...  is 3rd package");
@@ -166,6 +178,47 @@ filesavespackage(struct fileinlist *file,
   return true;
 }
 
+static void
+md5hash_prev_conffile(struct pkginfo *pkg, char *oldhash, const char *oldname,
+                      struct filenamenode *namenode)
+{
+  struct pkginfo *otherpkg;
+  struct conffile *conff;
+
+  debug(dbg_conffdetail, "tarobject looking for shared conffile %s",
+        namenode->name);
+
+  for (otherpkg = &pkg->set->pkg; otherpkg; otherpkg = otherpkg->arch_next) {
+    if (otherpkg == pkg)
+      continue;
+    /* The hash in the Conffiles is only meaningful if the package
+     * configuration has been at least tried. */
+    if (otherpkg->status <= stat_unpacked)
+      continue;
+    for (conff = otherpkg->installed.conffiles; conff; conff = conff->next) {
+      if (strcmp(conff->name, namenode->name) == 0)
+        break;
+    }
+    if (conff) {
+      debug(dbg_conffdetail,
+            "tarobject found shared conffile, from pkg %s (%s); hash=%s",
+            pkg_name(otherpkg, pnaw_always),
+            statusinfos[otherpkg->status].name, oldhash);
+      strcpy(oldhash, conff->hash);
+      break;
+    }
+  }
+
+  /* If no package was found with a valid Conffiles field, we make the
+   * risky assumption that the hash of the current .dpkg-new file is
+   * the one of the previously unpacked package. */
+  if (otherpkg == NULL) {
+    md5hash(pkg, oldhash, oldname);
+    debug(dbg_conffdetail,
+          "tarobject found shared conffile, from disk; hash=%s", oldhash);
+  }
+}
+
 void cu_pathname(int argc, void **argv) {
   ensure_pathname_nonexisting((char*)(argv[0]));
 }
@@ -181,11 +234,19 @@ int tarfileread(void *ud, char *buf, int len) {
 }
 
 static void
-tarfile_skip_one_forward(struct tarcontext *tc, struct tar_entry *ti)
+tarobject_skip_padding(struct tarcontext *tc, struct tar_entry *te)
 {
   size_t r;
-  char databuf[TARBLKSZ];
 
+  r = te->size % TARBLKSZ;
+  if (r > 0)
+    fd_skip(tc->backendpipe, TARBLKSZ - r,
+            _("error reading from dpkg-deb pipe"));
+}
+
+static void
+tarobject_skip_entry(struct tarcontext *tc, struct tar_entry *ti)
+{
   /* We need to advance the tar file to the next object, so read the
    * file data and set it to oblivion. */
   if (ti->type == tar_filetype_file) {
@@ -194,10 +255,7 @@ tarfile_skip_one_forward(struct tarcontext *tc, struct tar_entry *ti)
     fd_skip(tc->backendpipe, ti->size,
             _("skipped unpacking file '%.255s' (replaced or excluded?)"),
             path_quote_filename(fnamebuf, ti->name, 256));
-    r = ti->size % TARBLKSZ;
-    if (r > 0)
-      if (fd_read(tc->backendpipe, databuf, TARBLKSZ - r) < 0)
-        ohshite(_("error reading from dpkg-deb pipe"));
+    tarobject_skip_padding(tc, ti);
   }
 }
 
@@ -215,14 +273,21 @@ does_replace(struct pkginfo *newpigp, struct pkgbin *newpifp,
 {
   struct dependency *dep;
 
-  debug(dbg_depcon,"does_replace new=%s old=%s (%s)",newpigp->name,
-        oldpigp->name, versiondescribe(&oldpifp->version, vdew_always));
+  debug(dbg_depcon,"does_replace new=%s old=%s (%s)",
+        pkgbin_name(newpigp, newpifp, pnaw_always),
+        pkgbin_name(oldpigp, oldpifp, pnaw_always),
+        versiondescribe(&oldpifp->version, vdew_always));
   for (dep= newpifp->depends; dep; dep= dep->next) {
-    if (dep->type != dep_replaces || dep->list->ed != oldpigp) continue;
+    if (dep->type != dep_replaces || dep->list->ed != oldpigp->set)
+      continue;
     debug(dbg_depcondetail,"does_replace ... found old, version %s",
           versiondescribe(&dep->list->version,vdew_always));
     if (!versionsatisfied(oldpifp, dep->list))
       continue;
+    /* The test below can only trigger if dep_replaces start having
+     * arch qualifiers different from “any”. */
+    if (!archsatisfied(oldpifp, dep->list))
+      continue;
     debug(dbg_depcon,"does_replace ... yes");
     return true;
   }
@@ -230,6 +295,131 @@ does_replace(struct pkginfo *newpigp, struct pkgbin *newpifp,
   return false;
 }
 
+static void
+tarobject_extract(struct tarcontext *tc, struct tar_entry *te,
+                  const char *path, struct file_stat *st,
+                  struct filenamenode *namenode)
+{
+  static struct varbuf hardlinkfn;
+  static int fd;
+
+  struct filenamenode *linknode;
+  char fnamebuf[256];
+  char *newhash;
+
+  switch (te->type) {
+  case tar_filetype_file:
+    /* We create the file with mode 0 to make sure nobody can do anything with
+     * it until we apply the proper mode, which might be a statoverride. */
+    fd = open(path, O_CREAT | O_EXCL | O_WRONLY, 0);
+    if (fd < 0)
+      ohshite(_("unable to create `%.255s' (while processing `%.255s')"),
+              path, te->name);
+    push_cleanup(cu_closefd, ehflag_bombout, NULL, 0, 1, &fd);
+    debug(dbg_eachfiledetail, "tarobject file open size=%jd",
+          (intmax_t)te->size);
+
+    newhash = nfmalloc(MD5HASHLEN + 1);
+    fd_fd_copy_and_md5(tc->backendpipe, fd, newhash, te->size,
+                       _("backend dpkg-deb during `%.255s'"),
+                       path_quote_filename(fnamebuf, te->name, 256));
+    namenode->newhash = newhash;
+    debug(dbg_eachfiledetail, "tarobject file hash=%s", namenode->newhash);
+
+    tarobject_skip_padding(tc, te);
+
+    fd_writeback_init(fd);
+
+    if (namenode->statoverride)
+      debug(dbg_eachfile, "tarobject ... stat override, uid=%d, gid=%d, mode=%04o",
+            namenode->statoverride->uid,
+            namenode->statoverride->gid,
+            namenode->statoverride->mode);
+    if (fchown(fd, st->uid, st->gid))
+      ohshite(_("error setting ownership of `%.255s'"), te->name);
+    if (fchmod(fd, st->mode & ~S_IFMT))
+      ohshite(_("error setting permissions of `%.255s'"), te->name);
+
+    /* Postpone the fsync, to try to avoid massive I/O degradation. */
+    if (!fc_unsafe_io)
+      namenode->flags |= fnnf_deferred_fsync;
+
+    pop_cleanup(ehflag_normaltidy); /* fd = open(path) */
+    if (close(fd))
+      ohshite(_("error closing/writing `%.255s'"), te->name);
+    break;
+  case tar_filetype_fifo:
+    if (mkfifo(path, 0))
+      ohshite(_("error creating pipe `%.255s'"), te->name);
+    debug(dbg_eachfiledetail, "tarobject fifo");
+    break;
+  case tar_filetype_chardev:
+    if (mknod(path, S_IFCHR, te->dev))
+      ohshite(_("error creating device `%.255s'"), te->name);
+    debug(dbg_eachfiledetail, "tarobject chardev");
+    break;
+  case tar_filetype_blockdev:
+    if (mknod(path, S_IFBLK, te->dev))
+      ohshite(_("error creating device `%.255s'"), te->name);
+    debug(dbg_eachfiledetail, "tarobject blockdev");
+    break;
+  case tar_filetype_hardlink:
+    varbuf_reset(&hardlinkfn);
+    varbuf_add_str(&hardlinkfn, instdir);
+    varbuf_add_char(&hardlinkfn, '/');
+    linknode = findnamenode(te->linkname, 0);
+    varbuf_add_str(&hardlinkfn,
+                   namenodetouse(linknode, tc->pkg, &tc->pkg->available)->name);
+    if (linknode->flags & (fnnf_deferred_rename | fnnf_new_conff))
+      varbuf_add_str(&hardlinkfn, DPKGNEWEXT);
+    varbuf_end_str(&hardlinkfn);
+    if (link(hardlinkfn.buf, path))
+      ohshite(_("error creating hard link `%.255s'"), te->name);
+    namenode->newhash = linknode->newhash;
+    debug(dbg_eachfiledetail, "tarobject hardlink hash=%s", namenode->newhash);
+    break;
+  case tar_filetype_symlink:
+    /* We've already checked for an existing directory. */
+    if (symlink(te->linkname, path))
+      ohshite(_("error creating symbolic link `%.255s'"), te->name);
+    debug(dbg_eachfiledetail, "tarobject symlink creating");
+    break;
+  case tar_filetype_dir:
+    /* We've already checked for an existing directory. */
+    if (mkdir(path, 0))
+      ohshite(_("error creating directory `%.255s'"), te->name);
+    debug(dbg_eachfiledetail, "tarobject directory creating");
+    break;
+  default:
+    internerr("unknown tar type '%d', but already checked", te->type);
+  }
+}
+
+static void
+tarobject_hash(struct tarcontext *tc, struct tar_entry *te,
+               struct filenamenode *namenode)
+{
+  if (te->type == tar_filetype_file) {
+    char fnamebuf[256];
+    char *newhash;
+
+    newhash = nfmalloc(MD5HASHLEN + 1);
+    fd_md5(tc->backendpipe, newhash, te->size,
+           _("backend dpkg-deb during `%.255s'"),
+           path_quote_filename(fnamebuf, te->name, 256));
+    tarobject_skip_padding(tc, te);
+
+    namenode->newhash = newhash;
+    debug(dbg_eachfiledetail, "tarobject file hash=%s", namenode->newhash);
+  } else if (te->type == tar_filetype_hardlink) {
+    struct filenamenode *linknode;
+
+    linknode = findnamenode(te->linkname, 0);
+    namenode->newhash = linknode->newhash;
+    debug(dbg_eachfiledetail, "tarobject hardlink hash=%s", namenode->newhash);
+  }
+}
+
 static void
 tarobject_set_mtime(struct tar_entry *te, const char *path)
 {
@@ -307,6 +497,73 @@ tarobject_set_se_context(const char *matchpath, const char *path, mode_t mode)
 #endif /* WITH_SELINUX */
 }
 
+static void
+tarobject_matches(struct tarcontext *tc,
+                  const char *fn_old, struct stat *stab, char *oldhash,
+                  const char *fn_new, struct tar_entry *te,
+                  struct filenamenode *namenode)
+{
+  char *linkname;
+  ssize_t linksize;
+
+  debug(dbg_eachfiledetail, "tarobject matches on-disk object?");
+
+  switch (te->type) {
+  case tar_filetype_dir:
+    /* Nothing to check for a new directory. */
+    return;
+  case tar_filetype_symlink:
+    /* Symlinks to existing dirs have already been dealt with, only
+     * reamin real symlinks where we can compare the target. */
+    if (!S_ISLNK(stab->st_mode))
+      break;
+    linkname = m_malloc(stab->st_size + 1);
+    linksize = readlink(fn_old, linkname, stab->st_size + 1);
+    if (linksize < 0)
+      ohshite(_("unable to read link `%.255s'"), fn_old);
+    else if (linksize != stab->st_size)
+      ohshit(_("symbolic link '%.250s' size has changed from %jd to %zd"),
+             fn_old, stab->st_size, linksize);
+    linkname[linksize] = '\0';
+    if (strcmp(linkname, te->linkname) == 0) {
+      free(linkname);
+      return;
+    } else {
+      free(linkname);
+    }
+    break;
+  case tar_filetype_chardev:
+    if (S_ISCHR(stab->st_mode) && stab->st_rdev == te->dev)
+      return;
+    break;
+  case tar_filetype_blockdev:
+    if (S_ISBLK(stab->st_mode) && stab->st_rdev == te->dev)
+      return;
+    break;
+  case tar_filetype_fifo:
+    if (S_ISFIFO(stab->st_mode))
+      return;
+    break;
+  case tar_filetype_hardlink:
+    /* Fall through. */
+  case tar_filetype_file:
+    /* Only check metadata for non-conffiles. */
+    if (!(namenode->flags & fnnf_new_conff) &&
+        !(S_ISREG(stab->st_mode) && te->size == stab->st_size))
+      break;
+    if (strcmp(oldhash, namenode->newhash) == 0)
+      return;
+    break;
+  default:
+    internerr("unknown tar type '%d', but already checked", te->type);
+  }
+
+  forcibleerr(fc_overwrite,
+              _("trying to overwrite shared '%.250s', which is different "
+                "from other instances of package %.250s"),
+              namenode->name, pkg_name(tc->pkg, pnaw_nonambig));
+}
+
 void setupfnamevbs(const char *filename) {
   varbuf_trunc(&fnamevb, fnameidlu);
   varbuf_add_str(&fnamevb, filename);
@@ -429,22 +686,22 @@ linktosameexistingdir(const struct tar_entry *ti, const char *fname,
 int
 tarobject(void *ctx, struct tar_entry *ti)
 {
-  static struct varbuf conffderefn, hardlinkfn, symlinkfn;
-  static int fd;
+  static struct varbuf conffderefn, symlinkfn;
   const char *usename;
   struct filenamenode *usenode;
-  struct filenamenode *linknode;
 
   struct conffile *conff;
   struct tarcontext *tc = ctx;
   bool existingdir, keepexisting;
+  bool refcounting;
+  char oldhash[MD5HASHLEN + 1];
   int statr;
   ssize_t r;
   struct stat stab, stabtmp;
-  char databuf[TARBLKSZ];
   struct file_stat *st;
   struct fileinlist *nifd, **oldnifd;
-  struct pkginfo *divpkg, *otherpkg;
+  struct pkgset *divpkgset;
+  struct pkginfo *otherpkg;
 
   ensureobstackinit();
 
@@ -468,14 +725,14 @@ tarobject(void *ctx, struct tar_entry *ti)
         ? nifd->namenode->divert->useinstead->name : "<none>");
 
   if (nifd->namenode->divert && nifd->namenode->divert->camefrom) {
-    divpkg= nifd->namenode->divert->pkg;
+    divpkgset = nifd->namenode->divert->pkgset;
 
-    if (divpkg) {
+    if (divpkgset) {
       forcibleerr(fc_overwritediverted,
                   _("trying to overwrite `%.250s', which is the "
                     "diverted version of `%.250s' (package: %.100s)"),
                   nifd->namenode->name, nifd->namenode->divert->camefrom->name,
-                  divpkg->name);
+                  divpkgset->name);
     } else {
       forcibleerr(fc_overwritediverted,
                   _("trying to overwrite `%.250s', which is the "
@@ -489,7 +746,7 @@ tarobject(void *ctx, struct tar_entry *ti)
   else
     st = &ti->stat;
 
-  usenode = namenodetouse(nifd->namenode, tc->pkg);
+  usenode = namenodetouse(nifd->namenode, tc->pkg, &tc->pkg->available);
   usename = usenode->name + 1; /* Skip the leading '/'. */
 
   trig_file_activate(usenode, tc->pkg);
@@ -563,6 +820,7 @@ tarobject(void *ctx, struct tar_entry *ti)
   }
 
   keepexisting = false;
+  refcounting = false;
   if (!existingdir) {
     struct filepackages_iterator *iter;
 
@@ -570,16 +828,32 @@ tarobject(void *ctx, struct tar_entry *ti)
     while ((otherpkg = filepackages_iter_next(iter))) {
       if (otherpkg == tc->pkg)
         continue;
-      debug(dbg_eachfile, "tarobject ... found in %s", otherpkg->name);
+      debug(dbg_eachfile, "tarobject ... found in %s",
+            pkg_name(otherpkg, pnaw_always));
+
+      /* A pkgset can share files between its instances. Overwriting
+       * is allowed when they are not getting in sync, otherwise the
+       * file content must match the installed file. */
+      if (otherpkg->set == tc->pkg->set &&
+          otherpkg->installed.multiarch == multiarch_same &&
+          tc->pkg->available.multiarch == multiarch_same) {
+        if (statr == 0 && tc->pkgset_getting_in_sync)
+          refcounting = true;
+        debug(dbg_eachfiledetail, "tarobject ... shared with %s %s (syncing=%d)",
+              pkg_name(otherpkg, pnaw_always),
+              versiondescribe(&otherpkg->installed.version, vdew_nonambig),
+              tc->pkgset_getting_in_sync);
+        continue;
+      }
 
       if (nifd->namenode->divert && nifd->namenode->divert->useinstead) {
         /* Right, so we may be diverting this file. This makes the conflict
          * OK iff one of us is the diverting package (we don't need to
          * check for both being the diverting package, obviously). */
-        divpkg = nifd->namenode->divert->pkg;
-        debug(dbg_eachfile, "tarobject ... diverted, divpkg=%s",
-              divpkg ? divpkg->name : "<none>");
-        if (otherpkg == divpkg || tc->pkg == divpkg)
+        divpkgset = nifd->namenode->divert->pkgset;
+        debug(dbg_eachfile, "tarobject ... diverted, divpkgset=%s",
+              divpkgset ? divpkgset->name : "<none>");
+        if (otherpkg->set == divpkgset || tc->pkg->set == divpkgset)
           continue;
       }
 
@@ -638,12 +912,13 @@ tarobject(void *ctx, struct tar_entry *ti)
 
       if (does_replace(tc->pkg, &tc->pkg->available,
                        otherpkg, &otherpkg->installed)) {
-        printf(_("Replacing files in old package %s ...\n"),otherpkg->name);
+        printf(_("Replacing files in old package %s ...\n"),
+               pkg_name(otherpkg, pnaw_nonambig));
         otherpkg->clientdata->replacingfilesandsaid = 1;
       } else if (does_replace(otherpkg, &otherpkg->installed,
                               tc->pkg, &tc->pkg->available)) {
         printf(_("Replaced by files in installed package %s ...\n"),
-               otherpkg->name);
+               pkg_name(otherpkg, pnaw_nonambig));
         otherpkg->clientdata->replacingfilesandsaid = 2;
         nifd->namenode->flags &= ~fnnf_new_inarchive;
         keepexisting = true;
@@ -653,14 +928,14 @@ tarobject(void *ctx, struct tar_entry *ti)
           forcibleerr(fc_overwritedir,
                       _("trying to overwrite directory '%.250s' "
                         "in package %.250s %.250s with nondirectory"),
-                      nifd->namenode->name, otherpkg->name,
+                      nifd->namenode->name, pkg_name(otherpkg, pnaw_nonambig),
                       versiondescribe(&otherpkg->installed.version,
                                       vdew_nonambig));
         } else {
           forcibleerr(fc_overwrite,
                       _("trying to overwrite '%.250s', "
                         "which is also in package %.250s %.250s"),
-                      nifd->namenode->name, otherpkg->name,
+                      nifd->namenode->name, pkg_name(otherpkg, pnaw_nonambig),
                       versiondescribe(&otherpkg->installed.version,
                                       vdew_nonambig));
         }
@@ -671,14 +946,14 @@ tarobject(void *ctx, struct tar_entry *ti)
 
   if (keepexisting) {
     remove_file_from_list(tc, ti, oldnifd, nifd);
-    tarfile_skip_one_forward(tc, ti);
+    tarobject_skip_entry(tc, ti);
     return 0;
   }
 
   if (filter_should_skip(ti)) {
     nifd->namenode->flags &= ~fnnf_new_inarchive;
     nifd->namenode->flags |= fnnf_filtered;
-    tarfile_skip_one_forward(tc, ti);
+    tarobject_skip_entry(tc, ti);
 
     return 0;
   }
@@ -686,107 +961,51 @@ tarobject(void *ctx, struct tar_entry *ti)
   if (existingdir)
     return 0;
 
-  /* Now, at this stage we want to make sure neither of .dpkg-new and
-   * .dpkg-tmp are hanging around. */
-  ensure_pathname_nonexisting(fnamenewvb.buf);
-  ensure_pathname_nonexisting(fnametmpvb.buf);
-
-  /* Now we start to do things that we need to be able to undo
-   * if something goes wrong. Watch out for the CLEANUP comments to
-   * keep an eye on what's installed on the disk at each point. */
-  push_cleanup(cu_installnew, ~ehflag_normaltidy, NULL, 0, 1, (void *)nifd);
-
-  /*
-   * CLEANUP: Now we either have the old file on the disk, or not, in
-   * its original filename.
-   */
-
-  /* Extract whatever it is as .dpkg-new ... */
-  switch (ti->type) {
-  case tar_filetype_file:
-    /* We create the file with mode 0 to make sure nobody can do anything with
-     * it until we apply the proper mode, which might be a statoverride. */
-    fd= open(fnamenewvb.buf, (O_CREAT|O_EXCL|O_WRONLY), 0);
-    if (fd < 0)
-      ohshite(_("unable to create `%.255s' (while processing `%.255s')"),
-              fnamenewvb.buf, ti->name);
-    push_cleanup(cu_closefd, ehflag_bombout, NULL, 0, 1, &fd);
-    debug(dbg_eachfiledetail, "tarobject file open size=%jd",
-          (intmax_t)ti->size);
-    { char fnamebuf[256];
-    fd_fd_copy(tc->backendpipe, fd, ti->size,
-               _("backend dpkg-deb during `%.255s'"),
-               path_quote_filename(fnamebuf, ti->name, 256));
+  /* Compute the hash of the previous object, before we might replace it
+   * with the new version on forced overwrites. */
+  if (refcounting) {
+    if (nifd->namenode->flags & fnnf_new_conff) {
+      md5hash_prev_conffile(tc->pkg, oldhash, fnamenewvb.buf, nifd->namenode);
+    } else if (S_ISREG(stab.st_mode)) {
+      md5hash(tc->pkg, oldhash, fnamevb.buf);
+    } else {
+      strcpy(oldhash, EMPTYHASHFLAG);
     }
-    r = ti->size % TARBLKSZ;
-    if (r > 0)
-      if (fd_read(tc->backendpipe, databuf, TARBLKSZ - r) < 0)
-        ohshite(_("error reading from dpkg-deb pipe"));
-
-    fd_writeback_init(fd);
-
-    if (nifd->namenode->statoverride)
-      debug(dbg_eachfile, "tarobject ... stat override, uid=%d, gid=%d, mode=%04o",
-			  nifd->namenode->statoverride->uid,
-			  nifd->namenode->statoverride->gid,
-			  nifd->namenode->statoverride->mode);
-    if (fchown(fd, st->uid, st->gid))
-      ohshite(_("error setting ownership of `%.255s'"), ti->name);
-    if (fchmod(fd, st->mode & ~S_IFMT))
-      ohshite(_("error setting permissions of `%.255s'"), ti->name);
-
-    /* Postpone the fsync, to try to avoid massive I/O degradation. */
-    if (!fc_unsafe_io)
-      nifd->namenode->flags |= fnnf_deferred_fsync;
-
-    pop_cleanup(ehflag_normaltidy); /* fd = open(fnamenewvb.buf) */
-    if (close(fd))
-      ohshite(_("error closing/writing `%.255s'"), ti->name);
-    break;
-  case tar_filetype_fifo:
-    if (mkfifo(fnamenewvb.buf,0))
-      ohshite(_("error creating pipe `%.255s'"), ti->name);
-    debug(dbg_eachfiledetail, "tarobject fifo");
-    break;
-  case tar_filetype_chardev:
-    if (mknod(fnamenewvb.buf, S_IFCHR, ti->dev))
-      ohshite(_("error creating device `%.255s'"), ti->name);
-    debug(dbg_eachfiledetail, "tarobject chardev");
-    break;
-  case tar_filetype_blockdev:
-    if (mknod(fnamenewvb.buf, S_IFBLK, ti->dev))
-      ohshite(_("error creating device `%.255s'"), ti->name);
-    debug(dbg_eachfiledetail, "tarobject blockdev");
-    break;
-  case tar_filetype_hardlink:
-    varbuf_reset(&hardlinkfn);
-    varbuf_add_str(&hardlinkfn, instdir);
-    varbuf_add_char(&hardlinkfn, '/');
-    linknode = findnamenode(ti->linkname, 0);
-    varbuf_add_str(&hardlinkfn, namenodetouse(linknode, tc->pkg)->name);
-    if (linknode->flags & (fnnf_deferred_rename|fnnf_new_conff))
-      varbuf_add_str(&hardlinkfn, DPKGNEWEXT);
-    varbuf_end_str(&hardlinkfn);
-    if (link(hardlinkfn.buf,fnamenewvb.buf))
-      ohshite(_("error creating hard link `%.255s'"), ti->name);
-    debug(dbg_eachfiledetail, "tarobject hardlink");
-    break;
-  case tar_filetype_symlink:
-    /* We've already checked for an existing directory. */
-    if (symlink(ti->linkname, fnamenewvb.buf))
-      ohshite(_("error creating symbolic link `%.255s'"), ti->name);
-    debug(dbg_eachfiledetail, "tarobject symlink creating");
-    break;
-  case tar_filetype_dir:
-    /* We've already checked for an existing directory. */
-    if (mkdir(fnamenewvb.buf,0))
-      ohshite(_("error creating directory `%.255s'"), ti->name);
-    debug(dbg_eachfiledetail, "tarobject directory creating");
-    break;
-  default:
-    internerr("unknown tar type '%d', but already checked", ti->type);
   }
 
+  if (refcounting && !fc_overwrite) {
+    /* If we are not forced to overwrite the path and are refcounting,
+     * just compute the hash w/o extracting the object. */
+    tarobject_hash(tc, ti, nifd->namenode);
+  } else {
+    /* Now, at this stage we want to make sure neither of .dpkg-new and
+     * .dpkg-tmp are hanging around. */
+    ensure_pathname_nonexisting(fnamenewvb.buf);
+    ensure_pathname_nonexisting(fnametmpvb.buf);
+
+    /* Now we start to do things that we need to be able to undo
+     * if something goes wrong. Watch out for the CLEANUP comments to
+     * keep an eye on what's installed on the disk at each point. */
+    push_cleanup(cu_installnew, ~ehflag_normaltidy, NULL, 0, 1, nifd->namenode);
+
+    /*
+     * CLEANUP: Now we either have the old file on the disk, or not, in
+     * its original filename.
+     */
+
+    /* Extract whatever it is as .dpkg-new ... */
+    tarobject_extract(tc, ti, fnamenewvb.buf, st, nifd->namenode);
+  }
+
+  /* For shared files, check now if the object matches. */
+  if (refcounting)
+    tarobject_matches(tc, fnamevb.buf, &stab, oldhash,
+                          fnamenewvb.buf, ti, nifd->namenode);
+
+  /* If we didn't extract anything, there's nothing else to do. */
+  if (refcounting && !fc_overwrite)
+    return 0;
+
   tarobject_set_perms(ti, fnamenewvb.buf, st);
   tarobject_set_mtime(ti, fnamenewvb.buf);
   tarobject_set_se_context(fnamevb.buf, fnamenewvb.buf, st->mode);
@@ -892,7 +1111,7 @@ tar_writeback_barrier(struct fileinlist *files, struct pkginfo *pkg)
     if (!(cfile->namenode->flags & fnnf_deferred_fsync))
       continue;
 
-    usenode = namenodetouse(cfile->namenode, pkg);
+    usenode = namenodetouse(cfile->namenode, pkg, &pkg->available);
     usename = usenode->name + 1; /* Skip the leading '/'. */
 
     setupfnamevbs(usename);
@@ -930,7 +1149,7 @@ tar_deferred_extract(struct fileinlist *files, struct pkginfo *pkg)
     if (!(cfile->namenode->flags & fnnf_deferred_rename))
       continue;
 
-    usenode = namenodetouse(cfile->namenode, pkg);
+    usenode = namenodetouse(cfile->namenode, pkg, &pkg->available);
     usename = usenode->name + 1; /* Skip the leading '/'. */
 
     setupfnamevbs(usename);
@@ -973,6 +1192,20 @@ tar_deferred_extract(struct fileinlist *files, struct pkginfo *pkg)
   }
 }
 
+void
+enqueue_deconfigure(struct pkginfo *pkg, struct pkginfo *pkg_removal)
+{
+  struct pkg_deconf_list *newdeconf;
+
+  ensure_package_clientdata(pkg);
+  pkg->clientdata->istobe = itb_deconfigure;
+  newdeconf = m_malloc(sizeof(struct pkg_deconf_list));
+  newdeconf->next = deconfigure;
+  newdeconf->pkg = pkg;
+  newdeconf->pkg_removal = pkg_removal;
+  deconfigure = newdeconf;
+}
+
 /**
  * Try if we can deconfigure the package and queue it if so.
  *
@@ -994,8 +1227,6 @@ try_deconfigure_can(bool (*force_p)(struct deppossi *), struct pkginfo *pkg,
                     struct deppossi *pdep, const char *action,
                     struct pkginfo *removal, const char *why)
 {
-  struct pkg_deconf_list *newdeconf;
-
   if (force_p && force_p(pdep)) {
     warning(_("ignoring dependency problem with %s:\n%s"), action, why);
     return 2;
@@ -1003,20 +1234,16 @@ try_deconfigure_can(bool (*force_p)(struct deppossi *), struct pkginfo *pkg,
     if (pkg->installed.essential) {
       if (fc_removeessential) {
         warning(_("considering deconfiguration of essential\n"
-                  " package %s, to enable %s."), pkg->name, action);
+                  " package %s, to enable %s."),
+                pkg_name(pkg, pnaw_nonambig), action);
       } else {
         fprintf(stderr, _("dpkg: no, %s is essential, will not deconfigure\n"
                           " it in order to enable %s.\n"),
-                pkg->name, action);
+                pkg_name(pkg, pnaw_nonambig), action);
         return 0;
       }
     }
-    pkg->clientdata->istobe= itb_deconfigure;
-    newdeconf = m_malloc(sizeof(struct pkg_deconf_list));
-    newdeconf->next= deconfigure;
-    newdeconf->pkg= pkg;
-    newdeconf->pkg_removal = removal;
-    deconfigure= newdeconf;
+    enqueue_deconfigure(pkg, removal);
     return 1;
   } else {
     fprintf(stderr, _("dpkg: no, cannot proceed with %s (--auto-deconfigure will help):\n%s"),
@@ -1029,7 +1256,7 @@ static int try_remove_can(struct deppossi *pdep,
                           struct pkginfo *fixbyrm,
                           const char *why) {
   char action[512];
-  sprintf(action, _("removal of %.250s"), fixbyrm->name);
+  sprintf(action, _("removal of %.250s"), pkg_name(fixbyrm, pnaw_nonambig));
   return try_deconfigure_can(force_depends, pdep->up->up, pdep,
                              action, fixbyrm, why);
 }
@@ -1054,20 +1281,23 @@ void check_breaks(struct dependency *dep, struct pkginfo *pkg,
     ensure_package_clientdata(fixbydeconf);
     assert(fixbydeconf->clientdata->istobe == itb_normal);
 
-    sprintf(action, _("installation of %.250s"), pkg->name);
+    sprintf(action, _("installation of %.250s"),
+            pkgbin_name(pkg, &pkg->available, pnaw_nonambig));
     fprintf(stderr, _("dpkg: considering deconfiguration of %s,"
                       " which would be broken by %s ...\n"),
-            fixbydeconf->name, action);
+            pkg_name(fixbydeconf, pnaw_nonambig), action);
 
     ok= try_deconfigure_can(force_breaks, fixbydeconf, dep->list,
                             action, NULL, why.buf);
     if (ok == 1) {
       fprintf(stderr, _("dpkg: yes, will deconfigure %s (broken by %s).\n"),
-              fixbydeconf->name, pkg->name);
+              pkg_name(fixbydeconf, pnaw_nonambig),
+              pkgbin_name(pkg, &pkg->available, pnaw_nonambig));
     }
   } else {
     fprintf(stderr, _("dpkg: regarding %s containing %s:\n%s"),
-            pfilename, pkg->name, why.buf);
+            pfilename, pkgbin_name(pkg, &pkg->available, pnaw_nonambig),
+            why.buf);
     ok= 0;
   }
   varbuf_destroy(&why);
@@ -1081,10 +1311,11 @@ void check_breaks(struct dependency *dep, struct pkginfo *pkg,
   if (fixbydeconf && !f_autodeconf) {
     ohshit(_("installing %.250s would break %.250s, and\n"
              " deconfiguration is not permitted (--auto-deconfigure might help)"),
-           pkg->name, fixbydeconf->name);
+           pkgbin_name(pkg, &pkg->available, pnaw_nonambig),
+           pkg_name(fixbydeconf, pnaw_nonambig));
   } else {
     ohshit(_("installing %.250s would break existing software"),
-           pkg->name);
+           pkgbin_name(pkg, &pkg->available, pnaw_nonambig));
   }
 }
 
@@ -1114,16 +1345,17 @@ void check_conflict(struct dependency *dep, struct pkginfo *pkg,
       assert(fixbyrm->clientdata->istobe == itb_normal || fixbyrm->clientdata->istobe == itb_deconfigure);
       fixbyrm->clientdata->istobe= itb_remove;
       fprintf(stderr, _("dpkg: considering removing %s in favour of %s ...\n"),
-              fixbyrm->name, pkg->name);
+              pkg_name(fixbyrm, pnaw_nonambig),
+              pkgbin_name(pkg, &pkg->available, pnaw_nonambig));
       if (!(fixbyrm->status == stat_installed ||
             fixbyrm->status == stat_triggerspending ||
             fixbyrm->status == stat_triggersawaited)) {
         fprintf(stderr,
                 _("%s is not properly installed - ignoring any dependencies on it.\n"),
-                fixbyrm->name);
+                pkg_name(fixbyrm, pnaw_nonambig));
         pdep = NULL;
       } else {
-        for (pdep= fixbyrm->installed.depended;
+        for (pdep = fixbyrm->set->depended.installed;
              pdep;
              pdep = pdep->rev_next) {
           if (pdep->up->type != dep_depends && pdep->up->type != dep_predepends)
@@ -1140,7 +1372,7 @@ void check_conflict(struct dependency *dep, struct pkginfo *pkg,
                providecheck;
                providecheck= providecheck->next) {
             if (providecheck->type != dep_provides) continue;
-            for (pdep= providecheck->list->ed->installed.depended;
+            for (pdep = providecheck->list->ed->depended.installed;
                  pdep;
                  pdep = pdep->rev_next) {
               if (pdep->up->type != dep_depends && pdep->up->type != dep_predepends)
@@ -1150,7 +1382,8 @@ void check_conflict(struct dependency *dep, struct pkginfo *pkg,
               varbuf_end_str(&removalwhy);
               fprintf(stderr, _("dpkg"
                       ": may have trouble removing %s, as it provides %s ...\n"),
-                      fixbyrm->name, providecheck->list->ed->name);
+                      pkg_name(fixbyrm, pnaw_nonambig),
+                      providecheck->list->ed->name);
               if (!try_remove_can(pdep,fixbyrm,removalwhy.buf))
                 goto break_from_both_loops_at_once;
             }
@@ -1164,19 +1397,21 @@ void check_conflict(struct dependency *dep, struct pkginfo *pkg,
       if (!pdep && (fixbyrm->eflag & eflag_reinstreq)) {
         if (fc_removereinstreq) {
           fprintf(stderr, _("dpkg: package %s requires reinstallation, but will"
-                  " remove anyway as you requested.\n"), fixbyrm->name);
+                  " remove anyway as you requested.\n"),
+                  pkg_name(fixbyrm, pnaw_nonambig));
         } else {
           fprintf(stderr, _("dpkg: package %s requires reinstallation, "
-                  "will not remove.\n"), fixbyrm->name);
+                  "will not remove.\n"), pkg_name(fixbyrm, pnaw_nonambig));
           pdep= &flagdeppossi;
         }
       }
       if (!pdep) {
         /* This conflict is OK - we'll remove the conflictor. */
-        push_conflictor(pkg, fixbyrm);
+        enqueue_conflictor(pkg, fixbyrm);
         varbuf_destroy(&conflictwhy); varbuf_destroy(&removalwhy);
         fprintf(stderr, _("dpkg: yes, will remove %s in favour of %s.\n"),
-                fixbyrm->name, pkg->name);
+                pkg_name(fixbyrm, pnaw_nonambig),
+                pkgbin_name(pkg, &pkg->available, pnaw_nonambig));
         return;
       }
       /* Put it back. */
@@ -1185,9 +1420,11 @@ void check_conflict(struct dependency *dep, struct pkginfo *pkg,
   }
   varbuf_end_str(&conflictwhy);
   fprintf(stderr, _("dpkg: regarding %s containing %s:\n%s"),
-          pfilename, pkg->name, conflictwhy.buf);
+          pfilename, pkgbin_name(pkg, &pkg->available, pnaw_nonambig),
+          conflictwhy.buf);
   if (!force_conflicts(dep->list))
-    ohshit(_("conflicting packages - not installing %.250s"),pkg->name);
+    ohshit(_("conflicting packages - not installing %.250s"),
+           pkgbin_name(pkg, &pkg->available, pnaw_nonambig));
   warning(_("ignoring conflict, may proceed anyway!"));
   varbuf_destroy(&conflictwhy);
 
@@ -1221,6 +1458,8 @@ archivefiles(const char *const *argv)
                  msdbrw_available_write);
 
   checkpath();
+  pkg_infodb_upgrade();
+
   log_message("startup archives %s", cipaction->olong);
 
   if (f_recursive) {
@@ -1364,11 +1603,13 @@ wanttoinstall(struct pkginfo *pkg)
 
   if (pkg->want != want_install && pkg->want != want_hold) {
     if (f_alsoselect) {
-      printf(_("Selecting previously unselected package %s.\n"), pkg->name);
-      pkg->want = want_install;
+      printf(_("Selecting previously unselected package %s.\n"),
+             pkgbin_name(pkg, &pkg->available, pnaw_nonambig));
+      pkg_set_want(pkg, want_install);
       return true;
     } else {
-      printf(_("Skipping unselected package %s.\n"), pkg->name);
+      printf(_("Skipping unselected package %s.\n"),
+             pkgbin_name(pkg, &pkg->available, pnaw_nonambig));
       return false;
     }
   }
@@ -1387,20 +1628,22 @@ wanttoinstall(struct pkginfo *pkg)
       fprintf(stderr, _("Version %.250s of %.250s already installed, "
                         "skipping.\n"),
               versiondescribe(&pkg->installed.version, vdew_nonambig),
-              pkg->name);
+              pkg_name(pkg, pnaw_nonambig));
       return false;
     } else {
       return true;
     }
   } else {
     if (fc_downgrade) {
-      warning(_("downgrading %.250s from %.250s to %.250s."), pkg->name,
+      warning(_("downgrading %.250s from %.250s to %.250s."),
+              pkg_name(pkg, pnaw_nonambig),
               versiondescribe(&pkg->installed.version, vdew_nonambig),
               versiondescribe(&pkg->available.version, vdew_nonambig));
       return true;
     } else {
       fprintf(stderr, _("Will not downgrade %.250s from version %.250s "
-                        "to %.250s, skipping.\n"), pkg->name,
+                        "to %.250s, skipping.\n"),
+              pkg_name(pkg, pnaw_nonambig),
               versiondescribe(&pkg->installed.version, vdew_nonambig),
               versiondescribe(&pkg->available.version, vdew_nonambig));
       return false;
