@@ -4,6 +4,9 @@
  *
  * Copyright © 1995 Ian Jackson <ian@chiark.greenend.org.uk>
  * Copyright © 1999, 2002 Wichert Akkerman <wichert@deephackmode.org>
+ * Copyright © 2007-2012 Guillem Jover <guillem@debian.org>
+ * Copyright © 2011 Linaro Limited
+ * Copyright © 2011 Raphaël Hertzog <hertzog@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -42,6 +45,7 @@
 #include <dpkg/i18n.h>
 #include <dpkg/dpkg.h>
 #include <dpkg/dpkg-db.h>
+#include <dpkg/pkg.h>
 #include <dpkg/string.h>
 #include <dpkg/buffer.h>
 #include <dpkg/file.h>
@@ -59,7 +63,6 @@ static int conffoptcells[2][2] = {
 	{ cfo_keep,		cfo_prompt_keep },	/* User edited. */
 };
 
-static void md5hash(struct pkginfo *pkg, char *hashbuf, const char *fn);
 static void showdiff(const char *old, const char *new);
 static enum conffopt promptconfaction(struct pkginfo *pkg, const char *cfgfile,
                                       const char *realold, const char *realnew,
@@ -70,7 +73,6 @@ static void
 deferred_configure_conffile(struct pkginfo *pkg, struct conffile *conff)
 {
 	struct filenamenode *usenode;
-	static const char EMPTY_HASH[] = "-";
 	char currenthash[MD5HASHLEN + 1], newdisthash[MD5HASHLEN + 1];
 	int useredited, distedited;
 	enum conffopt what;
@@ -79,11 +81,12 @@ deferred_configure_conffile(struct pkginfo *pkg, struct conffile *conff)
 	char *cdr2rest;
 	int r;
 
-	usenode = namenodetouse(findnamenode(conff->name, fnn_nocopy), pkg);
+	usenode = namenodetouse(findnamenode(conff->name, fnn_nocopy),
+                                pkg, &pkg->installed);
 
 	r = conffderef(pkg, &cdr, usenode->name);
 	if (r == -1) {
-		conff->hash = EMPTY_HASH;
+		conff->hash = EMPTYHASHFLAG;
 		return;
 	}
 	md5hash(pkg, currenthash, cdr.buf);
@@ -166,7 +169,8 @@ deferred_configure_conffile(struct pkginfo *pkg, struct conffile *conff)
 		strcpy(cdr2rest, DPKGOLDEXT);
 		if (unlink(cdr2.buf) && errno != ENOENT)
 			warning(_("%s: failed to remove old backup '%.250s': %s"),
-			        pkg->name, cdr2.buf, strerror(errno));
+			        pkg_name(pkg, pnaw_nonambig), cdr2.buf,
+			        strerror(errno));
 
 		varbuf_add_str(&cdr, DPKGDISTEXT);
 		varbuf_end_str(&cdr);
@@ -174,27 +178,32 @@ deferred_configure_conffile(struct pkginfo *pkg, struct conffile *conff)
 		trig_file_activate(usenode, pkg);
 		if (rename(cdr2.buf, cdr.buf))
 			warning(_("%s: failed to rename '%.250s' to '%.250s': %s"),
-			        pkg->name, cdr2.buf, cdr.buf, strerror(errno));
+			        pkg_name(pkg, pnaw_nonambig), cdr2.buf, cdr.buf,
+			        strerror(errno));
 		break;
 	case cfo_keep:
 		strcpy(cdr2rest, DPKGNEWEXT);
 		if (unlink(cdr2.buf))
 			warning(_("%s: failed to remove '%.250s': %s"),
-			        pkg->name, cdr2.buf, strerror(errno));
+			        pkg_name(pkg, pnaw_nonambig), cdr2.buf,
+			        strerror(errno));
 		break;
 	case cfo_install | cfof_backup:
 		strcpy(cdr2rest, DPKGDISTEXT);
 		if (unlink(cdr2.buf) && errno != ENOENT)
 			warning(_("%s: failed to remove old distributed version '%.250s': %s"),
-			        pkg->name, cdr2.buf, strerror(errno));
+			        pkg_name(pkg, pnaw_nonambig), cdr2.buf,
+			        strerror(errno));
 		strcpy(cdr2rest, DPKGOLDEXT);
 		if (unlink(cdr2.buf) && errno != ENOENT)
 			warning(_("%s: failed to remove '%.250s' (before overwrite): %s"),
-			        pkg->name, cdr2.buf, strerror(errno));
+			        pkg_name(pkg, pnaw_nonambig), cdr2.buf,
+			        strerror(errno));
 		if (!(what & cfof_userrmd))
 			if (link(cdr.buf, cdr2.buf))
 				warning(_("%s: failed to link '%.250s' to '%.250s': %s"),
-				        pkg->name, cdr.buf, cdr2.buf, strerror(errno));
+				        pkg_name(pkg, pnaw_nonambig), cdr.buf,
+				        cdr2.buf, strerror(errno));
 		/* Fall through. */
 	case cfo_install:
 		printf(_("Installing new version of config file %s ...\n"),
@@ -251,18 +260,45 @@ deferred_configure(struct pkginfo *pkg)
 {
 	struct varbuf aemsgs = VARBUF_INIT;
 	struct conffile *conff;
+	struct pkginfo *otherpkg;
 	int ok;
 
 	if (pkg->status == stat_notinstalled)
 		ohshit(_("no package named `%s' is installed, cannot configure"),
-		       pkg->name);
+		       pkg_name(pkg, pnaw_nonambig));
 	if (pkg->status == stat_installed)
 		ohshit(_("package %.250s is already installed and configured"),
-		       pkg->name);
+		       pkg_name(pkg, pnaw_nonambig));
 	if (pkg->status != stat_unpacked && pkg->status != stat_halfconfigured)
 		ohshit(_("package %.250s is not ready for configuration\n"
 		         " cannot configure (current status `%.250s')"),
-		       pkg->name, statusinfos[pkg->status].name);
+		       pkg_name(pkg, pnaw_nonambig),
+		       statusinfos[pkg->status].name);
+
+	for (otherpkg = &pkg->set->pkg; otherpkg; otherpkg = otherpkg->arch_next) {
+		if (otherpkg == pkg)
+			continue;
+		if (otherpkg->status <= stat_configfiles)
+			continue;
+
+		if (otherpkg->status < stat_unpacked)
+			ohshit(_("package %s cannot be configured because "
+			         "%s is not ready (current status '%s')"),
+			       pkg_name(pkg, pnaw_always),
+			       pkg_name(otherpkg, pnaw_always),
+			       statusinfos[otherpkg->status].name);
+
+		if (versioncompare(&pkg->installed.version,
+		                   &otherpkg->installed.version))
+			ohshit(_("package %s %s cannot be configured because "
+			         "%s is at a different version (%s)"),
+			       pkg_name(pkg, pnaw_always),
+			       versiondescribe(&pkg->installed.version,
+			                       vdew_nonambig),
+			       pkg_name(otherpkg, pnaw_always),
+			       versiondescribe(&otherpkg->installed.version,
+			                       vdew_nonambig));
+	}
 
 	if (dependtry > 1)
 		if (findbreakcycle(pkg))
@@ -292,14 +328,14 @@ deferred_configure(struct pkginfo *pkg)
 		varbuf_end_str(&aemsgs);
 		fprintf(stderr,
 		        _("dpkg: dependency problems prevent configuration of %s:\n%s"),
-		        pkg->name, aemsgs.buf);
+		        pkg_name(pkg, pnaw_nonambig), aemsgs.buf);
 		varbuf_destroy(&aemsgs);
 		ohshit(_("dependency problems - leaving unconfigured"));
 	} else if (aemsgs.used) {
 		varbuf_end_str(&aemsgs);
 		fprintf(stderr,
 		        _("dpkg: %s: dependency problems, but configuring anyway as you requested:\n%s"),
-		        pkg->name, aemsgs.buf);
+		        pkg_name(pkg, pnaw_nonambig), aemsgs.buf);
 	}
 	varbuf_destroy(&aemsgs);
 	sincenothing = 0;
@@ -309,14 +345,14 @@ deferred_configure(struct pkginfo *pkg)
 		            _("Package is in a very bad inconsistent state - you should\n"
 		              " reinstall it before attempting configuration."));
 
-	printf(_("Setting up %s (%s) ...\n"), pkg->name,
+	printf(_("Setting up %s (%s) ...\n"), pkg_name(pkg, pnaw_nonambig),
 	       versiondescribe(&pkg->installed.version, vdew_nonambig));
-	log_action("configure", pkg);
+	log_action("configure", pkg, &pkg->installed);
 
 	trig_activate_packageprocessing(pkg);
 
 	if (f_noact) {
-		pkg->status = stat_installed;
+		pkg_set_status(pkg, stat_installed);
 		pkg->clientdata->istobe = itb_normal;
 		return;
 	}
@@ -342,7 +378,7 @@ deferred_configure(struct pkginfo *pkg)
 		for (conff = pkg->installed.conffiles; conff; conff = conff->next)
 			deferred_configure_conffile(pkg, conff);
 
-		pkg->status = stat_halfconfigured;
+		pkg_set_status(pkg, stat_halfconfigured);
 	}
 
 	assert(pkg->status == stat_halfconfigured);
@@ -355,7 +391,7 @@ deferred_configure(struct pkginfo *pkg)
 	                                           vdew_nonambig) : "",
 	                           NULL);
 
-	pkg->eflag = eflag_ok;
+	pkg_reset_eflags(pkg);
 	pkg->trigpend_head = NULL;
 	post_postinst_tasks(pkg, stat_installed);
 }
@@ -395,7 +431,8 @@ conffderef(struct pkginfo *pkg, struct varbuf *result, const char *in)
 			if (errno != ENOENT)
 				warning(_("%s: unable to stat config file '%s'\n"
 				          " (= '%s'): %s"),
-				        pkg->name, in, result->buf, strerror(errno));
+				        pkg_name(pkg, pnaw_nonambig), in,
+				        result->buf, strerror(errno));
 			debug(dbg_conffdetail, "conffderef nonexistent");
 			return 0;
 		} else if (S_ISREG(stab.st_mode)) {
@@ -407,7 +444,9 @@ conffderef(struct pkginfo *pkg, struct varbuf *result, const char *in)
 			      loopprotect);
 			if (loopprotect++ >= 25) {
 				warning(_("%s: config file '%s' is a circular link\n"
-				          " (= '%s')"), pkg->name, in, result->buf);
+				          " (= '%s')"),
+				        pkg_name(pkg, pnaw_nonambig), in,
+				        result->buf);
 				return -1;
 			}
 
@@ -417,7 +456,8 @@ conffderef(struct pkginfo *pkg, struct varbuf *result, const char *in)
 			if (r < 0) {
 				warning(_("%s: unable to readlink conffile '%s'\n"
 				          " (= '%s'): %s"),
-				        pkg->name, in, result->buf, strerror(errno));
+				        pkg_name(pkg, pnaw_nonambig), in,
+				        result->buf, strerror(errno));
 				return -1;
 			} else if (r != stab.st_size) {
 				warning(_("symbolic link '%.250s' size has "
@@ -443,8 +483,8 @@ conffderef(struct pkginfo *pkg, struct varbuf *result, const char *in)
 				if (r < 0) {
 					warning(_("%s: conffile '%.250s' resolves to degenerate filename\n"
 					          " ('%s' is a symlink to '%s')"),
-					        pkg->name, in, result->buf,
-					        target.buf);
+					        pkg_name(pkg, pnaw_nonambig),
+					        in, result->buf, target.buf);
 					return -1;
 				}
 				if (result->buf[r] == '/')
@@ -458,7 +498,7 @@ conffderef(struct pkginfo *pkg, struct varbuf *result, const char *in)
 			varbuf_end_str(result);
 		} else {
 			warning(_("%s: conffile '%.250s' is not a plain file or symlink (= '%s')"),
-			        pkg->name, in, result->buf);
+			        pkg_name(pkg, pnaw_nonambig), in, result->buf);
 			return -1;
 		}
 	}
@@ -474,7 +514,7 @@ conffderef(struct pkginfo *pkg, struct varbuf *result, const char *in)
  * @param[out] hashbuf The buffer to store the generated hash.
  * @param[in] fn The filename.
  */
-static void
+void
 md5hash(struct pkginfo *pkg, char *hashbuf, const char *fn)
 {
 	static int fd;
@@ -489,9 +529,9 @@ md5hash(struct pkginfo *pkg, char *hashbuf, const char *fn)
 	} else if (errno == ENOENT) {
 		strcpy(hashbuf, NONEXISTENTFLAG);
 	} else {
-		warning(_("%s: unable to open conffile %s for hash: %s"),
-		        pkg->name, fn, strerror(errno));
-		strcpy(hashbuf, "-");
+		warning(_("%s: unable to open %s for hash: %s"),
+		        pkg_name(pkg, pnaw_nonambig), fn, strerror(errno));
+		strcpy(hashbuf, EMPTYHASHFLAG);
 	}
 }
 
