@@ -3,6 +3,9 @@
  * enquiry.c - status enquiry and listing options
  *
  * Copyright © 1995,1996 Ian Jackson <ian@chiark.greenend.org.uk>
+ * Copyright © 2006,2008-2012 Guillem Jover <guillem@debian.org>
+ * Copyright © 2011 Linaro Limited
+ * Copyright © 2011 Raphaël Hertzog <hertzog@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -39,6 +42,7 @@
 #include <dpkg/i18n.h>
 #include <dpkg/dpkg.h>
 #include <dpkg/dpkg-db.h>
+#include <dpkg/arch.h>
 #include <dpkg/pkg-show.h>
 #include <dpkg/options.h>
 
@@ -54,17 +58,25 @@ struct badstatinfo {
 static bool
 bsyn_reinstreq(struct pkginfo *pkg, const struct badstatinfo *bsi)
 {
-  return pkg->eflag &= eflag_reinstreq;
+  return pkg->eflag & eflag_reinstreq;
 }
 
 static bool
 bsyn_status(struct pkginfo *pkg, const struct badstatinfo *bsi)
 {
-  if (pkg->eflag &= eflag_reinstreq)
+  if (pkg->eflag & eflag_reinstreq)
     return false;
   return (int)pkg->status == bsi->value;
 }
 
+static bool
+bsyn_arch(struct pkginfo *pkg, const struct badstatinfo *bsi)
+{
+  if (pkg->status < stat_halfinstalled)
+    return false;
+  return pkg->installed.arch->type == (enum dpkg_arch_type)bsi->value;
+}
+
 static const struct badstatinfo badstatinfos[]= {
   {
     .yesno = bsyn_reinstreq,
@@ -108,6 +120,21 @@ static const struct badstatinfo badstatinfos[]= {
     "The following packages have been triggered, but the trigger processing\n"
     "has not yet been done.  Trigger processing can be requested using\n"
     "dselect or dpkg --configure --pending (or dpkg --triggers-only):\n")
+  }, {
+    .yesno = bsyn_arch,
+    .value = arch_none,
+    .explanation = N_("The following packages do not have an architecture:\n")
+  }, {
+    .yesno = bsyn_arch,
+    .value = arch_illegal,
+    .explanation = N_("The following packages have an illegal architecture:\n")
+  }, {
+    .yesno = bsyn_arch,
+    .value = arch_unknown,
+    .explanation = N_(
+    "The following packages have an unknown foreign architecture, which will\n"
+    "cause dependency issues on front-ends. This can be fixed by registering\n"
+    "the foreign architecture with dpkg --add-architecture:\n")
   }, {
     .yesno = NULL
   }
@@ -118,13 +145,13 @@ static void describebriefly(struct pkginfo *pkg) {
   const char *pdesc;
 
   maxl= 57;
-  l= strlen(pkg->name);
+  l= strlen(pkg->set->name);
   if (l>20) maxl -= (l-20);
 
-  pdesc = pkg_summary(pkg, &l);
+  pdesc = pkg_summary(pkg, &pkg->installed, &l);
   l = min(l, maxl);
 
-  printf(" %-20s %.*s\n",pkg->name,l,pdesc);
+  printf(" %-20s %.*s\n", pkg_name(pkg, pnaw_nonambig), l, pdesc);
 }
 
 int
@@ -144,7 +171,7 @@ audit(const char *const *argv)
     bool head = false;
 
     it = pkg_db_iter_new();
-    while ((pkg = pkg_db_iter_next(it))) {
+    while ((pkg = pkg_db_iter_next_pkg(it))) {
       if (!bsi->yesno(pkg,bsi)) continue;
       if (!head_running) {
         if (modstatdb_is_locked())
@@ -216,7 +243,7 @@ unpackchk(const char *const *argv)
   sectionentries = NULL;
   sects= 0;
   it = pkg_db_iter_new();
-  while ((pkg = pkg_db_iter_next(it))) {
+  while ((pkg = pkg_db_iter_next_pkg(it))) {
     if (!yettobeunpacked(pkg, &thissect)) continue;
     for (se= sectionentries; se && strcasecmp(thissect,se->name); se= se->next);
     if (!se) {
@@ -239,7 +266,7 @@ unpackchk(const char *const *argv)
 
   if (totalcount <= 12) {
     it = pkg_db_iter_new();
-    while ((pkg = pkg_db_iter_next(it))) {
+    while ((pkg = pkg_db_iter_next_pkg(it))) {
       if (!yettobeunpacked(pkg, NULL))
         continue;
       describebriefly(pkg);
@@ -252,12 +279,16 @@ unpackchk(const char *const *argv)
       width= 70-strlen(se->name)-strlen(buf);
       while (width > 59) { putchar(' '); width--; }
       it = pkg_db_iter_new();
-      while ((pkg = pkg_db_iter_next(it))) {
+      while ((pkg = pkg_db_iter_next_pkg(it))) {
+        const char *pkgname;
+
         if (!yettobeunpacked(pkg,&thissect)) continue;
         if (strcasecmp(thissect,se->name)) continue;
-        width -= strlen(pkg->name); width--;
+        pkgname = pkg_name(pkg, pnaw_nonambig);
+        width -= strlen(pkgname);
+        width--;
         if (width < 4) { printf(" ..."); break; }
-        printf(" %s",pkg->name);
+        printf(" %s", pkgname);
       }
       pkg_db_iter_free(it);
       putchar('\n');
@@ -293,7 +324,7 @@ assert_version_support(const char *const *argv,
 
   modstatdb_open(msdbrw_readonly);
 
-  pkg = pkg_db_find("dpkg");
+  pkg = pkg_db_find_singleton("dpkg");
   switch (pkg->status) {
   case stat_installed:
   case stat_triggerspending:
@@ -346,6 +377,14 @@ assertmulticonrep(const char *const *argv)
                                 _("multiple Conflicts and Replaces"));
 }
 
+int
+assertmultiarch(const char *const *argv)
+{
+  struct versionrevision version = { 0, "1.16.2", NULL };
+
+  return assert_version_support(argv, &version, _("multi-arch"));
+}
+
 /**
  * Print a single package which:
  *  (a) is the target of one or more relevant predependencies.
@@ -378,7 +417,7 @@ predeppackage(const char *const *argv)
 
   dep = NULL;
   it = pkg_db_iter_new();
-  while (!dep && (pkg = pkg_db_iter_next(it))) {
+  while (!dep && (pkg = pkg_db_iter_next_pkg(it))) {
     /* Ignore packages user doesn't want. */
     if (pkg->want != want_install)
       continue;
@@ -412,20 +451,32 @@ predeppackage(const char *const *argv)
     for (possi = dep->list, pkg = NULL;
          !pkg && possi;
          possi=possi->next) {
-      trypkg= possi->ed;
-      if (trypkg->files && versionsatisfied(&trypkg->available,possi)) {
-        if (trypkg->clientdata->istobe == itb_normal) { pkg= trypkg; break; }
-      }
-      if (possi->verrel != dvr_none) continue;
-      for (provider=possi->ed->available.depended;
-           !pkg && provider;
-           provider=provider->next) {
-        if (provider->up->type != dep_provides) continue;
-        trypkg= provider->up->up;
-        if (!trypkg->files)
+      struct deppossi_pkg_iterator *possi_iter;
+
+      possi_iter = deppossi_pkg_iter_new(possi, wpb_available);
+      while (!pkg && (trypkg = deppossi_pkg_iter_next(possi_iter))) {
+        if (trypkg->files && trypkg->clientdata->istobe == itb_normal &&
+            versionsatisfied(&trypkg->available, possi)) {
+          pkg = trypkg;
+          break;
+        }
+        if (possi->verrel != dvr_none)
           continue;
-        if (trypkg->clientdata->istobe == itb_normal) { pkg= trypkg; break; }
+        for (provider = possi->ed->depended.available;
+             !pkg && provider;
+             provider = provider->next) {
+          if (provider->up->type != dep_provides)
+            continue;
+          trypkg = provider->up->up;
+          if (!trypkg->files)
+            continue;
+          if (trypkg->clientdata->istobe == itb_normal) {
+            pkg = trypkg;
+            break;
+          }
+        }
       }
+      deppossi_pkg_iter_free(possi_iter);
     }
     if (!pkg) {
       varbuf_reset(&vb);
@@ -433,7 +484,8 @@ predeppackage(const char *const *argv)
       varbuf_end_str(&vb);
       fprintf(stderr, _("dpkg: cannot see how to satisfy pre-dependency:\n %s\n"),vb.buf);
       ohshit(_("cannot satisfy pre-dependencies for %.250s (wanted due to %.250s)"),
-             dep->up->name,startpkg->name);
+             pkgbin_name(dep->up, &dep->up->available, pnaw_nonambig),
+             pkgbin_name(startpkg, &startpkg->available, pnaw_nonambig));
     }
     pkg->clientdata->istobe= itb_preinstall;
     for (dep= pkg->available.depends; dep; dep= dep->next) {
@@ -459,7 +511,7 @@ printarch(const char *const *argv)
   if (*argv)
     badusage(_("--%s takes no arguments"), cipaction->olong);
 
-  printf("%s\n", native_arch);
+  printf("%s\n", dpkg_arch_get(arch_native)->name);
 
   m_output(stdout, _("<standard output>"));
 
@@ -474,6 +526,28 @@ printinstarch(const char *const *argv)
   return printarch(argv);
 }
 
+int
+print_foreign_arches(const char *const *argv)
+{
+  struct dpkg_arch *arch;
+
+  if (*argv)
+    badusage(_("--%s takes no arguments"), cipaction->olong);
+
+  dpkg_arch_load_list();
+
+  for (arch = dpkg_arch_get_list(); arch; arch = arch->next) {
+    if (arch->type != arch_foreign)
+      continue;
+
+    printf("%s\n", arch->name);
+  }
+
+  m_output(stdout, _("<standard output>"));
+
+  return 0;
+}
+
 int
 cmpversions(const char *const *argv)
 {
