@@ -3,6 +3,7 @@
  * help.c - various helper routines
  *
  * Copyright © 1995 Ian Jackson <ian@chiark.greenend.org.uk>
+ * Copyright © 2007-2012 Guillem Jover <guillem@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -36,6 +37,7 @@
 #include <dpkg/i18n.h>
 #include <dpkg/dpkg.h>
 #include <dpkg/dpkg-db.h>
+#include <dpkg/pkg.h>
 #include <dpkg/path.h>
 #include <dpkg/subproc.h>
 #include <dpkg/command.h>
@@ -55,7 +57,10 @@ const char *const statusstrings[]= {
   [stat_installed]       = N_("installed")
 };
 
-struct filenamenode *namenodetouse(struct filenamenode *namenode, struct pkginfo *pkg) {
+struct filenamenode *
+namenodetouse(struct filenamenode *namenode, struct pkginfo *pkg,
+              struct pkgbin *pkgbin)
+{
   struct filenamenode *r;
 
   if (!namenode->divert) {
@@ -64,17 +69,17 @@ struct filenamenode *namenodetouse(struct filenamenode *namenode, struct pkginfo
   }
 
   debug(dbg_eachfile,"namenodetouse namenode=`%s' pkg=%s",
-        namenode->name,pkg->name);
+        namenode->name, pkgbin_name(pkg, pkgbin, pnaw_always));
 
   r=
-    (namenode->divert->useinstead && namenode->divert->pkg != pkg)
+    (namenode->divert->useinstead && namenode->divert->pkgset != pkg->set)
       ? namenode->divert->useinstead : namenode;
 
   debug(dbg_eachfile,
         "namenodetouse ... useinstead=%s camefrom=%s pkg=%s return %s",
         namenode->divert->useinstead ? namenode->divert->useinstead->name : "<none>",
         namenode->divert->camefrom ? namenode->divert->camefrom->name : "<none>",
-        namenode->divert->pkg ? namenode->divert->pkg->name : "<none>",
+        namenode->divert->pkgset ? namenode->divert->pkgset->name : "<none>",
         r->name);
 
   return r;
@@ -152,11 +157,29 @@ ignore_depends(struct pkginfo *pkg)
   return false;
 }
 
+static bool
+ignore_depends_possi(struct deppossi *possi)
+{
+  struct deppossi_pkg_iterator *possi_iter;
+  struct pkginfo *pkg;
+
+  possi_iter = deppossi_pkg_iter_new(possi, wpb_installed);
+  while ((pkg = deppossi_pkg_iter_next(possi_iter))) {
+    if (ignore_depends(pkg)) {
+      deppossi_pkg_iter_free(possi_iter);
+      return true;
+    }
+  }
+  deppossi_pkg_iter_free(possi_iter);
+
+  return false;
+}
+
 bool
 force_depends(struct deppossi *possi)
 {
   return fc_depends ||
-         ignore_depends(possi->ed) ||
+         ignore_depends_possi(possi) ||
          ignore_depends(possi->up->up);
 }
 
@@ -164,7 +187,7 @@ bool
 force_breaks(struct deppossi *possi)
 {
   return fc_breaks ||
-         ignore_depends(possi->ed) ||
+         ignore_depends_possi(possi) ||
          ignore_depends(possi->up->up);
 }
 
@@ -215,10 +238,13 @@ void
 post_postinst_tasks(struct pkginfo *pkg, enum pkgstatus new_status)
 {
   if (new_status < stat_triggersawaited)
-    pkg->status = new_status;
+    pkg_set_status(pkg, new_status);
+  else if (pkg->trigaw.head)
+    pkg_set_status(pkg, stat_triggersawaited);
+  else if (pkg->trigpend_head)
+    pkg_set_status(pkg, stat_triggerspending);
   else
-    pkg->status = pkg->trigaw.head ? stat_triggersawaited :
-                  pkg->trigpend_head ? stat_triggerspending : stat_installed;
+    pkg_set_status(pkg, stat_installed);
 
   post_postinst_tasks_core(pkg);
 }
@@ -269,8 +295,8 @@ do_script(struct pkginfo *pkg, struct pkgbin *pif,
 
   pid = subproc_fork();
   if (pid == 0) {
-    if (setenv("DPKG_MAINTSCRIPT_PACKAGE", pkg->name, 1) ||
-        setenv("DPKG_MAINTSCRIPT_ARCH", pif->arch, 1) ||
+    if (setenv("DPKG_MAINTSCRIPT_PACKAGE", pkg->set->name, 1) ||
+        setenv("DPKG_MAINTSCRIPT_ARCH", pif->arch->name, 1) ||
         setenv("DPKG_MAINTSCRIPT_NAME", cmd->argv[0], 1) ||
         setenv("DPKG_RUNNING_VERSION", PACKAGE_VERSION, 1))
       ohshite(_("unable to setenv for maintainer script"));
@@ -296,7 +322,7 @@ vmaintainer_script_installed(struct pkginfo *pkg, const char *scriptname,
   struct stat stab;
   char buf[100];
 
-  scriptpath= pkgadminfile(pkg,scriptname);
+  scriptpath = pkgadminfile(pkg, &pkg->installed, scriptname);
   sprintf(buf, _("installed %s script"), desc);
 
   command_init(&cmd, scriptpath, buf);
@@ -398,7 +424,7 @@ int maintainer_script_alternative(struct pkginfo *pkg,
   struct stat stab;
   char buf[100];
 
-  oldscriptpath= pkgadminfile(pkg,scriptname);
+  oldscriptpath = pkgadminfile(pkg, &pkg->installed, scriptname);
   sprintf(buf, _("old %s script"), desc);
 
   command_init(&cmd, oldscriptpath, buf);
@@ -455,7 +481,7 @@ void clear_istobes(void) {
   struct pkginfo *pkg;
 
   it = pkg_db_iter_new();
-  while ((pkg = pkg_db_iter_next(it)) != NULL) {
+  while ((pkg = pkg_db_iter_next_pkg(it)) != NULL) {
     ensure_package_clientdata(pkg);
     pkg->clientdata->istobe= itb_normal;
     pkg->clientdata->replacingfilesandsaid= 0;
@@ -474,13 +500,13 @@ dir_has_conffiles(struct filenamenode *file, struct pkginfo *pkg)
   size_t namelen;
 
   debug(dbg_veryverbose, "dir_has_conffiles '%s' (from %s)", file->name,
-	pkg->name);
+        pkg_name(pkg, pnaw_always));
   namelen = strlen(file->name);
   for (conff= pkg->installed.conffiles; conff; conff= conff->next) {
       if (strncmp(file->name, conff->name, namelen) == 0 &&
           conff->name[namelen] == '/') {
 	debug(dbg_veryverbose, "directory %s has conffile %s from %s",
-	      file->name, conff->name, pkg->name);
+	      file->name, conff->name, pkg_name(pkg, pnaw_always));
 	return true;
       }
   }
@@ -499,12 +525,12 @@ dir_is_used_by_others(struct filenamenode *file, struct pkginfo *pkg)
   struct pkginfo *other_pkg;
 
   debug(dbg_veryverbose, "dir_is_used_by_others '%s' (except %s)", file->name,
-        pkg ? pkg->name : "<none>");
+        pkg ? pkg_name(pkg, pnaw_always) : "<none>");
 
   iter = filepackages_iter_new(file);
   while ((other_pkg = filepackages_iter_next(iter))) {
     debug(dbg_veryverbose, "dir_is_used_by_others considering %s ...",
-          other_pkg->name);
+          pkg_name(other_pkg, pnaw_always));
     if (other_pkg == pkg)
       continue;
 
@@ -528,7 +554,7 @@ dir_is_used_by_pkg(struct filenamenode *file, struct pkginfo *pkg,
   size_t namelen;
 
   debug(dbg_veryverbose, "dir_is_used_by_pkg '%s' (by %s)",
-        file->name, pkg ? pkg->name : "<none>");
+        file->name, pkg ? pkg_name(pkg, pnaw_always) : "<none>");
 
   namelen = strlen(file->name);
 
@@ -624,9 +650,12 @@ void ensure_pathname_nonexisting(const char *pathname) {
   subproc_wait_check(pid, "rm cleanup", 0);
 }
 
-void log_action(const char *action, struct pkginfo *pkg) {
-  log_message("%s %s %s %s", action, pkg->name,
+void
+log_action(const char *action, struct pkginfo *pkg, struct pkgbin *pkgbin)
+{
+  log_message("%s %s %s %s", action, pkgbin_name(pkg, pkgbin, pnaw_always),
 	      versiondescribe(&pkg->installed.version, vdew_nonambig),
 	      versiondescribe(&pkg->available.version, vdew_nonambig));
-  statusfd_send("processing: %s: %s", action, pkg->name);
+  statusfd_send("processing: %s: %s", action,
+                pkgbin_name(pkg, pkgbin, pnaw_nonambig));
 }
