@@ -4,7 +4,9 @@
  *
  * Copyright © 1995,1996 Ian Jackson <ian@chiark.greenend.org.uk>
  * Copyright © 2000,2001 Wichert Akkerman <wakkerma@debian.org>
- * Copyright © 2006-2011 Guillem Jover <guillem@debian.org>
+ * Copyright © 2006-2012 Guillem Jover <guillem@debian.org>
+ * Copyright © 2011 Linaro Limited
+ * Copyright © 2011 Raphaël Hertzog <hertzog@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -43,6 +45,7 @@
 #include <dpkg/dpkg.h>
 #include <dpkg/dpkg-db.h>
 #include <dpkg/pkg-array.h>
+#include <dpkg/pkg-spec.h>
 #include <dpkg/pkg-format.h>
 #include <dpkg/pkg-show.h>
 #include <dpkg/path.h>
@@ -52,7 +55,9 @@
 #include "infodb.h"
 #include "main.h"
 
-static const char* showformat		= "${Package}\t${Version}\n";
+static const char *showformat = "${binary:Package}\t${Version}\n";
+
+static int opt_loadavail = 0;
 
 static int getwidth(void) {
   int fd;
@@ -102,10 +107,10 @@ list_format_init(struct list_format *fmt, struct pkg_array *array)
     for (i = 0; i < array->n_pkgs; i++) {
       int plen, vlen, dlen;
 
-      plen = strlen(array->pkgs[i]->name);
+      plen = strlen(pkg_name(array->pkgs[i], pnaw_nonambig));
       vlen = strlen(versiondescribe(&array->pkgs[i]->installed.version,
                                     vdew_nonambig));
-      pkg_summary(array->pkgs[i], &dlen);
+      pkg_summary(array->pkgs[i], &array->pkgs[i]->installed, &dlen);
 
       if (plen > fmt->nw)
         fmt->nw = plen;
@@ -183,14 +188,14 @@ list1package(struct pkginfo *pkg, struct list_format *fmt, struct pkg_array *arr
   list_format_init(fmt, array);
   list_format_print_header(fmt);
 
-  pdesc = pkg_summary(pkg, &l);
+  pdesc = pkg_summary(pkg, &pkg->installed, &l);
   l = min(l, fmt->dw);
 
   printf(fmt->format,
-         "uihrp"[pkg->want],
-         "ncHUFWti"[pkg->status],
-         " R"[pkg->eflag],
-         pkg->name,
+         pkg_abbrev_want(pkg),
+         pkg_abbrev_status(pkg),
+         pkg_abbrev_eflag(pkg),
+         pkg_name(pkg, pnaw_nonambig),
          versiondescribe(&pkg->installed.version, vdew_nonambig),
          l, pdesc);
 }
@@ -204,13 +209,13 @@ listpackages(const char *const *argv)
   int failures = 0;
   struct list_format fmt;
 
-  if (!*argv)
+  if (!opt_loadavail)
     modstatdb_open(msdbrw_readonly);
   else
     modstatdb_open(msdbrw_readonly | msdbrw_available_readonly);
 
   pkg_array_init_from_db(&array);
-  pkg_array_sort(&array, pkg_sorter_by_name);
+  pkg_array_sort(&array, pkg_sorter_by_nonambig_name_arch);
 
   fmt.head = false;
   fmt.format[0] = '\0';
@@ -223,15 +228,21 @@ listpackages(const char *const *argv)
     }
   } else {
     int argc, ip, *found;
+    struct pkg_spec *ps;
 
     for (argc = 0; argv[argc]; argc++);
-    found = m_malloc(sizeof(int) * argc);
-    memset(found, 0, sizeof(int) * argc);
+    found = m_calloc(sizeof(int) * argc);
+
+    ps = m_malloc(sizeof(*ps) * argc);
+    for (ip = 0; ip < argc; ip++) {
+      pkg_spec_init(&ps[ip], psf_patterns | psf_arch_def_wildcard);
+      pkg_spec_parse(&ps[ip], argv[ip]);
+    }
 
     for (i = 0; i < array.n_pkgs; i++) {
       pkg = array.pkgs[i];
       for (ip = 0; ip < argc; ip++) {
-        if (!fnmatch(argv[ip], pkg->name, 0)) {
+        if (pkg_spec_match_pkg(&ps[ip], pkg, &pkg->installed)) {
           list1package(pkg, &fmt, &array);
           found[ip]++;
           break;
@@ -246,7 +257,11 @@ listpackages(const char *const *argv)
         fprintf(stderr, _("No packages found matching %s.\n"), argv[ip]);
         failures++;
       }
+      pkg_spec_destroy(&ps[ip]);
     }
+
+    free(ps);
+    free(found);
   }
 
   m_output(stdout, _("<standard output>"));
@@ -269,11 +284,11 @@ static int searchoutput(struct filenamenode *namenode) {
     const char *name_to = namenode->divert->useinstead ?
                           namenode->divert->useinstead->name : namenode->name;
 
-    if (namenode->divert->pkg) {
+    if (namenode->divert->pkgset) {
       printf(_("diversion by %s from: %s\n"),
-             namenode->divert->pkg->name, name_from);
+             namenode->divert->pkgset->name, name_from);
       printf(_("diversion by %s to: %s\n"),
-             namenode->divert->pkg->name, name_to);
+             namenode->divert->pkgset->name, name_to);
     } else {
       printf(_("local diversion from: %s\n"), name_from);
       printf(_("local diversion to: %s\n"), name_to);
@@ -285,7 +300,7 @@ static int searchoutput(struct filenamenode *namenode) {
   while ((pkg_owner = filepackages_iter_next(iter))) {
     if (found)
       fputs(", ", stdout);
-    fputs(pkg_owner->name, stdout);
+    fputs(pkg_name(pkg_owner, pnaw_nonambig), stdout);
     found++;
   }
   filepackages_iter_free(iter);
@@ -380,7 +395,12 @@ enqperpackage(const char *const *argv)
     modstatdb_open(msdbrw_readonly);
 
   while ((thisarg = *argv++) != NULL) {
-    pkg = pkg_db_find(thisarg);
+    struct dpkg_error err;
+
+    pkg = pkg_spec_parse_pkg(thisarg, &err);
+    if (pkg == NULL)
+      badusage(_("--%s needs a valid package name but '%.250s' is not: %s"),
+               cipaction->olong, thisarg, err.str);
 
     switch (cipaction->arg_int) {
     case act_status:
@@ -390,7 +410,9 @@ enqperpackage(const char *const *argv)
           !pkg->files &&
           pkg->want == want_unknown &&
           !pkg_is_informative(pkg, &pkg->installed)) {
-        fprintf(stderr,_("Package `%s' is not installed and no info is available.\n"),pkg->name);
+        fprintf(stderr,
+                _("Package `%s' is not installed and no info is available.\n"),
+                pkg_name(pkg, pnaw_nonambig));
         failures++;
       } else {
         writerecord(stdout, _("<standard output>"), pkg, &pkg->installed);
@@ -398,7 +420,8 @@ enqperpackage(const char *const *argv)
       break;
     case act_printavail:
       if (!pkg_is_informative(pkg, &pkg->available)) {
-        fprintf(stderr,_("Package `%s' is not available.\n"),pkg->name);
+        fprintf(stderr, _("Package `%s' is not available.\n"),
+                pkgbin_name(pkg, &pkg->available, pnaw_nonambig));
         failures++;
       } else {
         writerecord(stdout, _("<standard output>"), pkg, &pkg->available);
@@ -407,7 +430,8 @@ enqperpackage(const char *const *argv)
     case act_listfiles:
       switch (pkg->status) {
       case stat_notinstalled:
-        fprintf(stderr,_("Package `%s' is not installed.\n"),pkg->name);
+        fprintf(stderr, _("Package `%s' is not installed.\n"),
+                pkg_name(pkg, pnaw_nonambig));
         failures++;
         break;
       default:
@@ -415,21 +439,22 @@ enqperpackage(const char *const *argv)
         ensure_diversions();
         file= pkg->clientdata->files;
         if (!file) {
-          printf(_("Package `%s' does not contain any files (!)\n"),pkg->name);
+          printf(_("Package `%s' does not contain any files (!)\n"),
+                 pkg_name(pkg, pnaw_nonambig));
         } else {
           while (file) {
             namenode= file->namenode;
             puts(namenode->name);
             if (namenode->divert && !namenode->divert->camefrom) {
-              if (!namenode->divert->pkg)
+              if (!namenode->divert->pkgset)
 		printf(_("locally diverted to: %s\n"),
 		       namenode->divert->useinstead->name);
-              else if (pkg == namenode->divert->pkg)
+              else if (pkg->set == namenode->divert->pkgset)
 		printf(_("package diverts others to: %s\n"),
 		       namenode->divert->useinstead->name);
               else
 		printf(_("diverted by %s to: %s\n"),
-		       namenode->divert->pkg->name,
+		       namenode->divert->pkgset->name,
 		       namenode->divert->useinstead->name);
             }
             file= file->next;
@@ -472,13 +497,13 @@ showpackages(const char *const *argv)
     return failures;
   }
 
-  if (!*argv)
+  if (!opt_loadavail)
     modstatdb_open(msdbrw_readonly);
   else
     modstatdb_open(msdbrw_readonly | msdbrw_available_readonly);
 
   pkg_array_init_from_db(&array);
-  pkg_array_sort(&array, pkg_sorter_by_name);
+  pkg_array_sort(&array, pkg_sorter_by_nonambig_name_arch);
 
   if (!*argv) {
     for (i = 0; i < array.n_pkgs; i++) {
@@ -488,15 +513,21 @@ showpackages(const char *const *argv)
     }
   } else {
     int argc, ip, *found;
+    struct pkg_spec *ps;
 
     for (argc = 0; argv[argc]; argc++);
-    found = m_malloc(sizeof(int) * argc);
-    memset(found, 0, sizeof(int) * argc);
+    found = m_calloc(sizeof(int) * argc);
+
+    ps = m_malloc(sizeof(*ps) * argc);
+    for (ip = 0; ip < argc; ip++) {
+      pkg_spec_init(&ps[ip], psf_patterns | psf_arch_def_wildcard);
+      pkg_spec_parse(&ps[ip], argv[ip]);
+    }
 
     for (i = 0; i < array.n_pkgs; i++) {
       pkg = array.pkgs[i];
       for (ip = 0; ip < argc; ip++) {
-        if (!fnmatch(argv[ip], pkg->name, 0)) {
+        if (pkg_spec_match_pkg(&ps[ip], pkg, &pkg->installed)) {
           pkg_format_show(fmt, pkg, &pkg->installed);
           found[ip]++;
           break;
@@ -511,7 +542,11 @@ showpackages(const char *const *argv)
         fprintf(stderr, _("No packages found matching %s.\n"), argv[ip]);
         failures++;
       }
+      pkg_spec_destroy(&ps[ip]);
     }
+
+    free(ps);
+    free(found);
   }
 
   m_output(stdout, _("<standard output>"));
@@ -544,7 +579,7 @@ control_path_file(struct pkginfo *pkg, const char *control_file)
   const char *control_path;
   struct stat st;
 
-  control_path = pkgadminfile(pkg, control_file);
+  control_path = pkgadminfile(pkg, &pkg->installed, control_file);
   if (stat(control_path, &st) < 0)
     return;
   if (!S_ISREG(st.st_mode))
@@ -556,12 +591,13 @@ control_path_file(struct pkginfo *pkg, const char *control_file)
 static int
 control_path(const char *const *argv)
 {
+  struct dpkg_error err;
   struct pkginfo *pkg;
-  const char *pkg_name;
+  const char *pkgname;
   const char *control_file;
 
-  pkg_name = *argv++;
-  if (!pkg_name)
+  pkgname = *argv++;
+  if (!pkgname)
     badusage(_("--%s needs at least one package name argument"),
              cipaction->olong);
 
@@ -580,14 +616,19 @@ control_path(const char *const *argv)
 
   modstatdb_open(msdbrw_readonly);
 
-  pkg = pkg_db_find(pkg_name);
+  pkg = pkg_spec_parse_pkg(pkgname, &err);
+  if (pkg == NULL)
+    badusage(_("--%s needs a valid package name but '%.250s' is not: %s"),
+             cipaction->olong, pkgname, err.str);
+
   if (pkg->status == stat_notinstalled)
-    ohshit(_("Package `%s' is not installed.\n"), pkg->name);
+    ohshit(_("Package `%s' is not installed.\n"),
+           pkg_name(pkg, pnaw_nonambig));
 
   if (control_file)
     control_path_file(pkg, control_file);
   else
-    pkg_infodb_foreach(pkg, pkg_infodb_print_filename);
+    pkg_infodb_foreach(pkg, &pkg->installed, pkg_infodb_print_filename);
 
   modstatdb_shutdown();
 
@@ -635,6 +676,7 @@ usage(const struct cmdinfo *ci, const char *value)
   printf(_(
 "Options:\n"
 "  --admindir=<directory>           Use <directory> instead of %s.\n"
+"  --load-avail                     Use available file on --show and --list.\n"
 "  -f|--showformat=<format>         Use alternative format for --show.\n"
 "\n"), ADMINDIR);
 
@@ -670,6 +712,7 @@ static const struct cmdinfo cmdinfos[]= {
   ACTION( "control-path",                   'c', act_controlpath,   control_path    ),
 
   { "admindir",   0,   1, NULL, &admindir,   NULL          },
+  { "load-avail", 0,   0, &opt_loadavail, NULL, NULL, 1    },
   { "showformat", 'f', 1, NULL, &showformat, NULL          },
   { "help",       'h', 0, NULL, NULL,        usage         },
   { "version",    0,   0, NULL, NULL,        printversion  },
