@@ -3,6 +3,7 @@
  * remove.c - functionality for removing packages
  *
  * Copyright © 1995 Ian Jackson <ian@chiark.greenend.org.uk>
+ * Copyright © 2007-2012 Guillem Jover <guillem@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -36,6 +37,7 @@
 #include <dpkg/i18n.h>
 #include <dpkg/dpkg.h>
 #include <dpkg/dpkg-db.h>
+#include <dpkg/pkg.h>
 #include <dpkg/dir.h>
 #include <dpkg/options.h>
 #include <dpkg/triglib.h>
@@ -48,22 +50,24 @@
  * pkgdepcheck may be a virtual pkg.
  */
 static void checkforremoval(struct pkginfo *pkgtoremove,
-                            struct pkginfo *pkgdepcheck,
+                            struct pkgset *pkgdepcheck,
                             int *rokp, struct varbuf *raemsgs) {
   struct deppossi *possi;
   struct pkginfo *depender;
   int before, ok;
 
-  for (possi = pkgdepcheck->installed.depended; possi; possi = possi->rev_next) {
+  for (possi = pkgdepcheck->depended.installed; possi; possi = possi->rev_next) {
     if (possi->up->type != dep_depends && possi->up->type != dep_predepends) continue;
     depender= possi->up->up;
-    debug(dbg_depcon,"checking depending package `%s'",depender->name);
+    debug(dbg_depcon, "checking depending package `%s'",
+          pkg_name(depender, pnaw_always));
     if (!(depender->status == stat_installed ||
           depender->status == stat_triggerspending ||
           depender->status == stat_triggersawaited))
       continue;
     if (ignore_depends(depender)) {
-      debug(dbg_depcon, "ignoring depending package '%s'", depender->name);
+      debug(dbg_depcon, "ignoring depending package '%s'",
+            pkg_name(depender, pnaw_always));
       continue;
     }
     if (dependtry > 1) { if (findbreakcycle(pkgtoremove)) sincenothing= 0; }
@@ -82,11 +86,12 @@ void deferred_remove(struct pkginfo *pkg) {
   int rok;
   struct dependency *dep;
 
-  debug(dbg_general,"deferred_remove package %s",pkg->name);
+  debug(dbg_general, "deferred_remove package %s",
+        pkg_name(pkg, pnaw_always));
 
   if (pkg->status == stat_notinstalled) {
     warning(_("ignoring request to remove %.250s which isn't installed."),
-            pkg->name);
+            pkg_name(pkg, pnaw_nonambig));
     pkg->clientdata->istobe= itb_normal;
     return;
   } else if (!f_pending &&
@@ -94,7 +99,7 @@ void deferred_remove(struct pkginfo *pkg) {
              cipaction->arg_int != act_purge) {
     warning(_("ignoring request to remove %.250s, only the config\n"
               " files of which are on the system. Use --purge to remove them too."),
-            pkg->name);
+            pkg_name(pkg, pnaw_nonambig));
     pkg->clientdata->istobe= itb_normal;
     return;
   }
@@ -103,17 +108,22 @@ void deferred_remove(struct pkginfo *pkg) {
     forcibleerr(fc_removeessential, _("This is an essential package -"
                 " it should not be removed."));
 
-  if (!f_pending)
-    pkg->want = (cipaction->arg_int == act_purge) ? want_purge : want_deinstall;
+  if (!f_pending) {
+    if (cipaction->arg_int == act_purge)
+      pkg_set_want(pkg, want_purge);
+    else
+      pkg_set_want(pkg, want_deinstall);
+  }
   if (!f_noact) modstatdb_note(pkg);
 
-  debug(dbg_general,"checking dependencies for remove `%s'",pkg->name);
+  debug(dbg_general, "checking dependencies for remove `%s'",
+        pkg_name(pkg, pnaw_always));
   rok= 2;
-  checkforremoval(pkg,pkg,&rok,&raemsgs);
+  checkforremoval(pkg, pkg->set, &rok, &raemsgs);
   for (dep= pkg->installed.depends; dep; dep= dep->next) {
     if (dep->type != dep_provides) continue;
     debug(dbg_depcon,"checking virtual package `%s'",dep->list->ed->name);
-    checkforremoval(pkg,dep->list->ed,&rok,&raemsgs);
+    checkforremoval(pkg, dep->list->ed, &rok, &raemsgs);
   }
 
   if (rok == 1) {
@@ -126,13 +136,13 @@ void deferred_remove(struct pkginfo *pkg) {
     varbuf_end_str(&raemsgs);
     fprintf(stderr,
             _("dpkg: dependency problems prevent removal of %s:\n%s"),
-            pkg->name, raemsgs.buf);
+            pkg_name(pkg, pnaw_nonambig), raemsgs.buf);
     ohshit(_("dependency problems - not removing"));
   } else if (raemsgs.used) {
     varbuf_end_str(&raemsgs);
     fprintf(stderr,
             _("dpkg: %s: dependency problems, but removing anyway as you requested:\n%s"),
-            pkg->name, raemsgs.buf);
+            pkg_name(pkg, pnaw_nonambig), raemsgs.buf);
   }
   varbuf_destroy(&raemsgs);
   sincenothing= 0;
@@ -146,22 +156,22 @@ void deferred_remove(struct pkginfo *pkg) {
   filesdbinit();
 
   if (f_noact) {
-    printf(_("Would remove or purge %s ...\n"),pkg->name);
-    pkg->status= stat_notinstalled;
+    printf(_("Would remove or purge %s ...\n"), pkg_name(pkg, pnaw_nonambig));
+    pkg_set_status(pkg, stat_notinstalled);
     pkg->clientdata->istobe= itb_normal;
     return;
   }
 
   oldconffsetflags(pkg->installed.conffiles);
 
-  printf(_("Removing %s ...\n"),pkg->name);
-  log_action("remove", pkg);
+  printf(_("Removing %s ...\n"), pkg_name(pkg, pnaw_nonambig));
+  log_action("remove", pkg, &pkg->installed);
   trig_activate_packageprocessing(pkg);
   if (pkg->status >= stat_halfconfigured) {
     static enum pkgstatus oldpkgstatus;
 
     oldpkgstatus= pkg->status;
-    pkg->status= stat_halfconfigured;
+    pkg_set_status(pkg, stat_halfconfigured);
     modstatdb_note(pkg);
     push_cleanup(cu_prermremove, ~ehflag_normaltidy, NULL, 0, 2,
                  (void *)pkg, (void *)&oldpkgstatus);
@@ -169,7 +179,7 @@ void deferred_remove(struct pkginfo *pkg) {
                                 "remove", NULL);
 
     /* Will turn into ‘half-installed’ soon ... */
-    pkg->status = stat_unpacked;
+    pkg_set_status(pkg, stat_unpacked);
   }
 
   removal_bulk(pkg);
@@ -200,6 +210,33 @@ removal_bulk_remove_file(const char *filename, const char *filetype)
   debug(dbg_scripts, "removal_bulk info unlinked %s", filename);
 }
 
+static bool
+removal_bulk_file_is_shared(struct pkginfo *pkg, struct filenamenode *namenode)
+{
+  struct filepackages_iterator *iter;
+  struct pkginfo *otherpkg;
+  bool shared = false;
+
+  if (pkgset_installed_instances(pkg->set) <= 1)
+    return false;
+
+  iter = filepackages_iter_new(namenode);
+  while ((otherpkg = filepackages_iter_next(iter))) {
+    if (otherpkg == pkg)
+      continue;
+    if (otherpkg->set != pkg->set)
+      continue;
+
+    debug(dbg_eachfiledetail, "removal_bulk file shared with %s, skipping",
+          pkg_name(otherpkg, pnaw_always));
+    shared = true;
+    break;
+  }
+  filepackages_iter_free(iter);
+
+  return shared;
+}
+
 static void
 removal_bulk_remove_files(struct pkginfo *pkg)
 {
@@ -210,7 +247,7 @@ removal_bulk_remove_files(struct pkginfo *pkg)
   static struct varbuf fnvb;
   struct stat stab;
 
-    pkg->status= stat_halfinstalled;
+    pkg_set_status(pkg, stat_halfinstalled);
     modstatdb_note(pkg);
     push_checkpoint(~ehflag_bombout, ehflag_normaltidy);
 
@@ -218,22 +255,36 @@ removal_bulk_remove_files(struct pkginfo *pkg)
     leftover = NULL;
     while ((namenode= reversefilelist_next(&rlistit))) {
       struct filenamenode *usenode;
+      bool is_dir;
 
       debug(dbg_eachfile, "removal_bulk `%s' flags=%o",
             namenode->name, namenode->flags);
-      if (namenode->flags & fnnf_old_conff) {
-        push_leftover(&leftover,namenode);
-        continue;
-      }
 
-      usenode = namenodetouse(namenode, pkg);
-      trig_file_activate(usenode, pkg);
+      usenode = namenodetouse(namenode, pkg, &pkg->installed);
 
       varbuf_reset(&fnvb);
       varbuf_add_str(&fnvb, instdir);
       varbuf_add_str(&fnvb, usenode->name);
+      varbuf_end_str(&fnvb);
       before= fnvb.used;
 
+      is_dir = stat(fnvb.buf, &stab) == 0 && S_ISDIR(stab.st_mode);
+
+      /* A pkgset can share files between its instances that we
+       * don't want to remove, we just want to forget them. This
+       * applies to shared conffiles too. */
+      if (!is_dir && removal_bulk_file_is_shared(pkg, namenode))
+        continue;
+
+      /* Non-shared conffiles are kept. */
+      if (namenode->flags & fnnf_old_conff) {
+        push_leftover(&leftover, namenode);
+        continue;
+      }
+
+      trig_file_activate(usenode, pkg);
+
+      varbuf_trunc(&fnvb, before);
       varbuf_add_str(&fnvb, DPKGTEMPEXT);
       varbuf_end_str(&fnvb);
       debug(dbg_eachfiledetail, "removal_bulk cleaning temp `%s'", fnvb.buf);
@@ -248,7 +299,7 @@ removal_bulk_remove_files(struct pkginfo *pkg)
 
       varbuf_trunc(&fnvb, before);
       varbuf_end_str(&fnvb);
-      if (!stat(fnvb.buf,&stab) && S_ISDIR(stab.st_mode)) {
+      if (is_dir) {
         debug(dbg_eachfiledetail, "removal_bulk is a directory");
         /* Only delete a directory or a link to one if we're the only
          * package which uses it. Other files should only be listed
@@ -274,7 +325,7 @@ removal_bulk_remove_files(struct pkginfo *pkg)
       } else if (errno == EBUSY || errno == EPERM) {
         warning(_("while removing %.250s, unable to remove directory '%.250s': "
                   "%s - directory may be a mount point?"),
-                pkg->name, namenode->name, strerror(errno));
+                pkg_name(pkg, pnaw_nonambig), namenode->name, strerror(errno));
         push_leftover(&leftover,namenode);
         continue;
       } else if (errno == EINVAL && strcmp(usenode->name, "/.") == 0) {
@@ -288,19 +339,19 @@ removal_bulk_remove_files(struct pkginfo *pkg)
       if (secure_unlink(fnvb.buf))
         ohshite(_("unable to securely remove '%.250s'"), fnvb.buf);
     }
-    write_filelist_except(pkg,leftover,0);
+    write_filelist_except(pkg, &pkg->installed, leftover, 0);
     maintainer_script_installed(pkg, POSTRMFILE, "post-removal",
                                 "remove", NULL);
 
-    trig_parse_ci(pkgadminfile(pkg, TRIGGERSCIFILE),
-                  trig_cicb_interest_delete, NULL, pkg);
+    trig_parse_ci(pkgadminfile(pkg, &pkg->installed, TRIGGERSCIFILE),
+                  trig_cicb_interest_delete, NULL, pkg, &pkg->installed);
     trig_file_interests_save();
 
     debug(dbg_general, "removal_bulk cleaning info directory");
-    pkg_infodb_foreach(pkg, removal_bulk_remove_file);
+    pkg_infodb_foreach(pkg, &pkg->installed, removal_bulk_remove_file);
     dir_sync_path(pkgadmindir());
 
-    pkg->status= stat_configfiles;
+    pkg_set_status(pkg, stat_configfiles);
     pkg->installed.essential = false;
     modstatdb_note(pkg);
     push_checkpoint(~ehflag_bombout, ehflag_normaltidy);
@@ -334,7 +385,7 @@ static void removal_bulk_remove_leftover_dirs(struct pkginfo *pkg) {
       continue;
     }
 
-    usenode = namenodetouse(namenode, pkg);
+    usenode = namenodetouse(namenode, pkg, &pkg->installed);
     trig_file_activate(usenode, pkg);
 
     varbuf_reset(&fnvb);
@@ -359,13 +410,13 @@ static void removal_bulk_remove_leftover_dirs(struct pkginfo *pkg) {
     if (!rmdir(fnvb.buf) || errno == ENOENT || errno == ELOOP) continue;
     if (errno == ENOTEMPTY || errno == EEXIST) {
       warning(_("while removing %.250s, directory '%.250s' not empty so not removed."),
-              pkg->name, namenode->name);
+              pkg_name(pkg, pnaw_nonambig), namenode->name);
       push_leftover(&leftover,namenode);
       continue;
     } else if (errno == EBUSY || errno == EPERM) {
       warning(_("while removing %.250s, unable to remove directory '%.250s': "
                 "%s - directory may be a mount point?"),
-              pkg->name, namenode->name, strerror(errno));
+              pkg_name(pkg, pnaw_nonambig), namenode->name, strerror(errno));
       push_leftover(&leftover,namenode);
       continue;
     } else if (errno == EINVAL && strcmp(usenode->name, "/.") == 0) {
@@ -388,7 +439,7 @@ static void removal_bulk_remove_leftover_dirs(struct pkginfo *pkg) {
     push_leftover(&leftover,namenode);
     continue;
   }
-  write_filelist_except(pkg,leftover,0);
+  write_filelist_except(pkg, &pkg->installed, leftover, 0);
 
   modstatdb_note(pkg);
   push_checkpoint(~ehflag_bombout, ehflag_normaltidy);
@@ -406,8 +457,9 @@ static void removal_bulk_remove_configfiles(struct pkginfo *pkg) {
   char *p;
   const char *const *ext;
 
-    printf(_("Purging configuration files for %s ...\n"),pkg->name);
-    log_action("purge", pkg);
+    printf(_("Purging configuration files for %s ...\n"),
+           pkg_name(pkg, pnaw_nonambig));
+    log_action("purge", pkg, &pkg->installed);
     trig_activate_packageprocessing(pkg);
 
     /* We may have modified this above. */
@@ -431,7 +483,7 @@ static void removal_bulk_remove_configfiles(struct pkginfo *pkg) {
       } else if (searchfile->namenode->divert &&
                  (searchfile->namenode->divert->camefrom ||
                   (searchfile->namenode->divert->useinstead &&
-                   searchfile->namenode->divert->pkg != pkg))) {
+                   searchfile->namenode->divert->pkgset != pkg->set))) {
         debug(dbg_conff, "removal_bulk conffile '%s' ignored due to diversion",
               conff->name);
         *lconffp= conff->next;
@@ -514,7 +566,8 @@ static void removal_bulk_remove_configfiles(struct pkginfo *pkg) {
     }
 
     /* Remove the conffiles from the file list file. */
-    write_filelist_except(pkg, pkg->clientdata->files, fnnf_old_conff);
+    write_filelist_except(pkg, &pkg->installed, pkg->clientdata->files,
+                          fnnf_old_conff);
 
     pkg->installed.conffiles = NULL;
     modstatdb_note(pkg);
@@ -531,13 +584,13 @@ static void removal_bulk_remove_configfiles(struct pkginfo *pkg) {
 void removal_bulk(struct pkginfo *pkg) {
   bool foundpostrm;
 
-  debug(dbg_general,"removal_bulk package %s",pkg->name);
+  debug(dbg_general, "removal_bulk package %s", pkg_name(pkg, pnaw_always));
 
   if (pkg->status == stat_halfinstalled || pkg->status == stat_unpacked) {
     removal_bulk_remove_files(pkg);
   }
 
-  foundpostrm = pkg_infodb_has_file(pkg, POSTRMFILE);
+  foundpostrm = pkg_infodb_has_file(pkg, &pkg->installed, POSTRMFILE);
 
   debug(dbg_general, "removal_bulk purging? foundpostrm=%d",foundpostrm);
 
@@ -545,8 +598,8 @@ void removal_bulk(struct pkginfo *pkg) {
     /* If there are no config files and no postrm script then we
      * go straight into ‘purge’.  */
     debug(dbg_general, "removal_bulk no postrm, no conffiles, purging");
-    pkg->want= want_purge;
 
+    pkg_set_want(pkg, want_purge);
     blankversion(&pkg->configversion);
   } else if (pkg->want == want_purge) {
 
@@ -561,27 +614,27 @@ void removal_bulk(struct pkginfo *pkg) {
     /* Retry empty directories, and warn on any leftovers that aren't. */
     removal_bulk_remove_leftover_dirs(pkg);
 
-    filename = pkgadminfile(pkg, LISTFILE);
+    filename = pkgadminfile(pkg, &pkg->installed, LISTFILE);
     debug(dbg_general, "removal_bulk purge done, removing list `%s'",
           filename);
     if (unlink(filename) && errno != ENOENT)
       ohshite(_("cannot remove old files list"));
 
-    filename = pkgadminfile(pkg, POSTRMFILE);
+    filename = pkgadminfile(pkg, &pkg->installed, POSTRMFILE);
     debug(dbg_general, "removal_bulk purge done, removing postrm `%s'",
           filename);
     if (unlink(filename) && errno != ENOENT)
       ohshite(_("can't remove old postrm script"));
 
-    pkg->status= stat_notinstalled;
-    pkg->want = want_unknown;
+    pkg_set_status(pkg, stat_notinstalled);
+    pkg_set_want(pkg, want_unknown);
 
     /* This will mess up reverse links, but if we follow them
      * we won't go back because pkg->status is stat_notinstalled. */
     pkgbin_blank(&pkg->installed);
   }
 
-  pkg->eflag = eflag_ok;
+  pkg_reset_eflags(pkg);
   modstatdb_note(pkg);
 
   debug(dbg_general, "removal done");
