@@ -80,11 +80,6 @@ version(void)
 	printf(_("Debian %s version %s.\n"), PROGNAME, VERSION);
 	printf("\n");
 
-	printf(_(
-"Copyright (C) 1995 Ian Jackson.\n"
-"Copyright (C) 2000-2002 Wichert Akkerman.\n"
-"Copyright (C) 2009-2010 Raphael Hertzog.\n"));
-
 	printf(_(
 "This is free software; see the GNU General Public License version 2 or\n"
 "later for copying conditions. There is NO warranty.\n"));
@@ -2309,8 +2304,8 @@ main(int argc, char **argv)
 	int i = 0;
 
 	setlocale(LC_ALL, "");
-	bindtextdomain("dpkg", LOCALEDIR);
-	textdomain("dpkg");
+	bindtextdomain(PACKAGE, LOCALEDIR);
+	textdomain(PACKAGE);
 
 	admdir = admindir_init();
 
